import numpy as np
from ruleskit import HyperrectangleCondition, RegressionRule
from ruleskit import Rule
from ruleskit import RuleSet
import pytest


@pytest.mark.parametrize(
    "rule1, rule2, rs",
    [
        (
            Rule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
            Rule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
            RuleSet(
                [
                    Rule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
                    Rule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
                ]
            ),
        ),
    ],
)
def test_add(rule1, rule2, rs):
    res = RuleSet([rule1])
    res += RuleSet([rule2])
    np.testing.assert_equal(res, rs)


@pytest.mark.parametrize(
    "xs, y, rule_list, coverage",
    [
        (
            np.array([[5, 3], [3, 4], [2, np.nan]]),
            np.array([1, 3, 2]),
            [
                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
            ],
            2 / 3,
        ),
    ],
)
def test_coverage(clean, xs, y, rule_list, coverage):
    [r.fit(xs, y) for r in rule_list]
    res = RuleSet(rule_list, remember_activation=True, stack_activation=True)
    np.testing.assert_equal(res.ruleset_coverage, coverage)
    res.calc_activation(xs)
    np.testing.assert_equal(res.ruleset_coverage, coverage)


@pytest.mark.parametrize(
    "rs",
    [
        RuleSet(
            [
                Rule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
                Rule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
            ]
        ),
    ],
)
def test_save(rs):
    rs.save("tests/unit/data/ruleset.csv")


@pytest.mark.parametrize(
    "rs",
    [
        RuleSet(
            [
                Rule(HyperrectangleCondition([0, 1], bmins=[1, 1], bmaxs=[2, 3])),
                Rule(HyperrectangleCondition([1, 2], bmins=[4, 4], bmaxs=[5, 4])),
            ]
        ),
    ],
)
def test_save(rs):
    rs.save("tests/unit/data/ruleset.csv")


def test_load():
    rs = RuleSet()
    rs.load("tests/unit/data/ruleset_test.csv")
    for rule in rs:
        print(rule)



from abc import ABC
import ast
from copy import copy
from typing import List, Union, Tuple, Any, Optional
from collections import Counter
import numpy as np
import itertools
from collections import OrderedDict

from .condition import HyperrectangleCondition

from .rule import Rule, ClassificationRule, RegressionRule
from .activation import Activation
import logging

logger = logging.getLogger(__name__)


class RuleSet(ABC):

    """A set of rules"""

    NLINES = 5  # half how many rules to show in str(self)
    CHECK_DUPLICATED = False
    STACKED_FIT = False
    all_features_indexes = {}

    condition_index = ["features_names", "features_indexes", "bmins", "bmaxs"]
    rule_index = ["prediction"]
    classification_rule_index = rule_index + ["coverage", "criterion"]
    regression_rule_index = classification_rule_index + ["std"]

    @staticmethod
    def check_duplicated_rules(rules, name_or_index: str = "index"):
        if name_or_index == "index":
            str_rules = [str(r.features_indexes) + str(r.bmins) + str(r.bmaxs) + str(r.prediction) for r in rules]
        else:
            str_rules = [str(r.features_names) + str(r.bmins) + str(r.bmaxs) + str(r.prediction) for r in rules]
        if len(set(str_rules)) < len(str_rules):
            duplicated = {}
            for r in str_rules:
                if r not in duplicated:
                    duplicated[r] = 0
                duplicated[r] += 1
            duplicated = [
                f"{r}: {duplicated[r]} (positions {[i for i, x in enumerate(str_rules) if x == r]})\n"
                f"   underlying rules : {[str(rules[i]) for i in [i for i, x in enumerate(str_rules) if x == r]]}"
                for r in duplicated
                if duplicated[r] > 1
            ]
            s = "\n -".join(duplicated)
            raise ValueError(f"There are {len(duplicated)} duplicated rules in your ruleset !\n {s}")

    def __init__(
        self,
        rules_list: Union[List[Rule], None] = None,
        remember_activation: bool = True,
        stack_activation: bool = False,
    ):
        """

        Parameters
        ----------
        rules_list: Union[List[Rule], None]
            The list of rules to start with. Can be None, since a RuleSet can be filled after its creation.
        remember_activation: bool
            The activation of the RuleSet is the logical OR of the activation of all its rules. It is only computed
            if remember_activation is True. (default value = True)
        stack_activation: bool
            If True, the RuleSet will keep in memory 2-D np.ndarray containing the activations of all its rules. This
            can take a lot of memory, but can save time if you apply numpy methods on this stacked vector instead of on
            each rule separately. (default value = False)
        """
        self._rules = []
        self.features_names = []
        self.features_indexes = []
        self._activation = None
        self._coverage = None  # in case Activation is not available
        self.stacked_activations = None
        self.remember_activation = remember_activation
        self.stack_activation = stack_activation
        self.rule_type = None
        if rules_list is not None:
            names_available = all([hasattr(r.condition, "features_names") for r in self])
            for rule in rules_list:
                if not isinstance(rule, Rule) and rule is not None:
                    raise TypeError(f"Some rules in given iterable were not of type 'Rule' but of type {type(rule)}")
                if rule is not None:
                    self.append(rule, update_activation=False)
            if self.remember_activation:
                self.compute_self_activation()
            if self.stack_activation:
                self.compute_stacked_activation()
            if names_available:
                self.features_names = list(set(itertools.chain(*[rule.features_names for rule in self])))
            self.set_features_indexes()
        if RuleSet.CHECK_DUPLICATED:
            self.check_duplicated_rules(self.rules, name_or_index="name" if len(self.features_names) > 0 else "index")

    # noinspection PyProtectedMember,PyTypeChecker
    def __iadd__(self, other: Union["RuleSet", Rule]):
        """Appends a rule or each rules of another RuleSet to self and updates activation vector and stacked activations
        if needed. Also updates features_indexes, and features_names if possible."""
        if isinstance(other, Rule):
            self._rules.append(other)
        else:
            self._rules += other._rules
        self.features_indexes = list(set(self.features_indexes + other.features_indexes))
        if hasattr(other, "features_names"):
            self.features_names = list(set(self.features_names + other.features_names))
        if self.remember_activation:
            self._update_activation(other)
        if self.stack_activation:
            self._update_stacked_activation(other)
        return self

    def __add__(self, other: Union["RuleSet", Rule]):
        """Returns the RuleSet resulting in appendind a rule or each rules of another RuleSet to self."""
        remember_activation = self.remember_activation
        stack_activation = self.stack_activation
        if isinstance(other, Rule):
            rules = self.rules + [other]
        else:
            remember_activation &= other.remember_activation
            stack_activation &= other.stack_activation
            rules = list(set(self.rules + other.rules))
        return self.__class__(rules, remember_activation=remember_activation, stack_activation=stack_activation)

    def __getattr__(self, item):
        """If item is not found in self, try to fetch it from its activation."""
        if item == "_activation":
            raise AttributeError(f"'RuleSet' object has no attribute '{item}'")

        if self._activation is not None and hasattr(self._activation, item):
            return getattr(self._activation, item)
        raise AttributeError(f"'RuleSet' object has no attribute '{item}'")

    def __len__(self):
        """The length of a RuleSet its the number of rules stored in it."""
        return len(self.rules)

    def __eq__(self, other: "RuleSet"):
        return set(self.rules) == set(other.rules)

    def __iter__(self):
        if hasattr(self, "_rules"):
            return self.rules.__iter__()
        else:
            return [].__iter__()

    def __getitem__(self, key):
        if isinstance(key, slice):
            indices = range(*key.indices(len(self.rules)))
            return self.__class__([self.rules[i] for i in indices])
        return self.rules.__getitem__(key)

    def __str__(self):
        if len(self) < 2 * RuleSet.NLINES:
            return "\n".join([str(self[i]) for i in range(len(self))])
        else:
            return "\n".join(
                [str(self[i]) for i in range(RuleSet.NLINES)]
                + ["..."]
                + [str(self[i]) for i in range(len(self) - RuleSet.NLINES, len(self))]
            )

    def __hash__(self) -> hash:
        return hash(frozenset(self.to_hash))

    # noinspection PyProtectedMember
    def __contains__(self, other: Rule) -> bool:
        """A RuleSet contains another Rule if the two rule's conditions and predictions are the same"""
        name_or_index = "name" if len(self.features_names) > 0 else "index"

        if name_or_index == "index":
            str_rules = [str(r.features_indexes) + str(r.bmins) + str(r.bmaxs) + str(r.prediction) for r in self]
            str_rule = str(other.features_indexes) + str(other.bmins) + str(other.bmaxs) + str(other.prediction)
        else:
            str_rules = [str(r.features_names) + str(r.bmins) + str(r.bmaxs) + str(r.prediction) for r in self]
            str_rule = str(other.features_names) + str(other.bmins) + str(other.bmaxs) + str(other.prediction)

        return str_rule in str_rules

    @property
    def rules(self) -> List[Rule]:
        return self._rules

    @rules.setter
    def rules(self, rules: Union[List[Rule], None]):
        ruleset = RuleSet(rules, remember_activation=self.remember_activation, stack_activation=self.stack_activation)
        self._rules = ruleset._rules
        self.features_names = ruleset.features_names
        self.features_indexes = ruleset.features_indexes
        self.stacked_activations = ruleset.stacked_activations
        self._activation = ruleset._activation

    @property
    def to_hash(self) -> Tuple[str]:
        if len(self) == 0:
            return "rs",
        to_hash = ("rs",)
        for r in self:
            rule_hash = r.to_hash[1:]
            to_hash += rule_hash
        return to_hash

    @property
    def activation_available(self) -> bool:
        """Returns True if the RuleSet has an activation vector, and if this Activation's object data is available."""
        if self._activation is None:
            return False
        if self._activation.data_format == "file":
            return self._activation.data.is_file()
        else:
            return self._activation.data is not None

    @property
    def stacked_activations_available(self) -> bool:
        """Returns True is the RuleSet has its rules' stacked activations."""
        if self.stack_activation is None:
            return False
        return True

    @property
    def activation(self) -> Union[None, np.ndarray]:
        """Returns the Activation vector's data in a form of a 1-D np.ndarray, or None if not available.

        Returns:
        --------
        Union[None, np.ndarray]
            of the form [0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, ...]
        """
        if self._activation:
            return self._activation.raw
        return None

    @property
    def ruleset_coverage(self) -> float:
        """Coverage is the fraction of points equal to 1 in the activation vector"""
        if not self.activation_available:
            return self._coverage
        else:
            return self._activation.coverage

    # noinspection PyProtectedMember,PyTypeChecker
    def _update_activation(self, other: Union[Rule, "RuleSet"]):
        """Updates the activation vector of the RuleSet with the activation vector of a new Rule or RuleSet."""
        if other.activation_available:
            if self._activation is None:
                self._activation = Activation(other.activation, to_file=Rule.LOCAL_ACTIVATION)
            else:
                self._activation = self._activation | other._activation

    # noinspection PyProtectedMember,PyTypeChecker
    def _update_stacked_activation(self, other: Union[Rule, "RuleSet"]):
        """Updates the stacked activation vectors of the RuleSet with the activation vector of a new Rule or
        the stacked activation vectors of another RuleSet."""
        if other.activation_available:
            try:
                import pandas as pd
            except ImportError:
                raise ImportError("RuleSet's stacked activations requied pandas. Please run\npip install pandas")

            if self.stacked_activations is None:
                if isinstance(other, Rule):
                    self.stacked_activations = pd.DataFrame(
                        data=np.array(other.activation).T, columns=[str(other.condition)]
                    )
                else:
                    self.stacked_activations = other.stacked_activations
            else:
                if isinstance(other, Rule):
                    self.stacked_activations[str(other.condition)] = other.activation
                else:
                    self.stacked_activations = pd.concat([self.stacked_activations, other.stacked_activations], axis=1)

    def set_features_indexes(self):
        if len(RuleSet.all_features_indexes) > 0:
            self.features_indexes = [RuleSet.all_features_indexes[f] for f in self.features_names]
            for r in self._rules:
                # noinspection PyProtectedMember
                r._condition._features_indexes = [RuleSet.all_features_indexes[f] for f in r.features_names]
        else:
            list(set(itertools.chain(*[rule.features_indexes for rule in self])))

    # noinspection PyProtectedMember
    def fit(
        self, y: Union[np.ndarray, "pd.Series"], xs: Optional[Union["pd.DataFrame", np.ndarray]] = None
    ) -> List[Rule]:
        if len(self) == 0:
            logger.debug("Ruleset is empty. Nothing to fit.")
            return []

        if RuleSet.STACKED_FIT:
            try:
                import pandas as pd
            except ImportError:
                raise ImportError("RuleSet's stacked fit requies pandas. Please run\npip install pandas")

            clean_activation = False
            if xs is not None:
                clean_activation = not self.stack_activation
                self.stack_activation = True
                self.calc_activation(xs)
                self.stack_activation = not clean_activation
            elif self.stacked_activations is None:
                clean_activation = not self.stack_activation
                self.stack_activation = True
                self.compute_stacked_activation()
                self.stack_activation = not clean_activation

            if isinstance(y, np.ndarray) and not len(self.stacked_activations.index) == y.shape[0]:
                raise IndexError(
                    "Stacked activation and y have different number of rows. Use pd.Series for y to"
                    " reindex stacked activations automatically."
                )
            else:
                self.stacked_activations.index = y.index
            pred, prediction_vector = self.calc_prediction(y)

            if clean_activation:
                self.del_stacked_activations()
            else:
                self.stacked_activations.index = pd.RangeIndex(len(y.index))

            prediction_vector.index = self.stack_activations.index
            criterion = self.calc_criterion(prediction_vector, y)
            to_drop = []

            for ir in range(len(self)):
                self._rules[ir]._criterion = criterion.iloc[ir] if isinstance(criterion, pd.Series) else criterion
                self._rules[ir]._prediction = pred.iloc[ir] if isinstance(pred, pd.Series) else pred

                self._rules[ir].check_thresholds("prediction")
                if not self._rules[ir].good:
                    to_drop.append(self._rules[ir])
                    break
        else:
            [r.fit(xs=xs, y=y) for r in self]
            to_drop = [r for r in self if not r.good]

        if len(to_drop) > 0:
            rules = [r for r in self.rules if r not in to_drop]
            self._rules = []
            if self._activation is not None:
                self._activation.clear()
            self.del_stacked_activations()
            for r in rules:
                self.append(r)
        return to_drop

    def append(self, rule: Rule, update_activation: bool = True):
        """Appends a new rule to self. The updates of activation vector and the stacked activation vectors can be
        blocked by specifying update_activation=False. Otherwise, will use self.remember_activation and
        self.stack_activation to determine if the updates should be done or not."""
        if not isinstance(rule, Rule):
            raise TypeError(f"RuleSet's append method expects a Rule object, got {type(rule)}")
        if self.rule_type is None:
            self.rule_type = type(rule)
        else:
            if not isinstance(rule, self.rule_type):
                raise TypeError(
                    f"Ruleset previously had rules of type {self.rule_type}, so can not add rule of type {type(rule)}"
                )
        remember_activation = self.remember_activation
        stack_activation = self.stack_activation
        if not update_activation:
            self.remember_activation = False
            self.stack_activation = False
        self.__iadd__(rule)
        self.remember_activation = remember_activation
        self.stack_activation = stack_activation

    def sort(self, criterion: str = None, reverse: bool = False):
        """Sorts the RuleSet.

        * If criterion is not speficied:
            Will sort the rules according to :
                1. The number of features they talk about
                2. For a same number of features (sorted in alphabetical order, or index if names are not available,
                    optionally reversed), the bmins and bmaxs of the rules
        * If criterion is specified, it must be an float or interger attribute of rule, condition or activation. Then
            sorts according to this criterion, optionally reversed.
        """
        if len(self) == 0:
            return

        if criterion is None or criterion == "":
            if not (hasattr(self[0].condition, "bmins") and hasattr(self[0].condition, "bmaxs")):
                return
            # The set of all the features the RuleSet talks about
            which = "index"
            if len(self.features_names) > 0:
                which = "name"
                fnames_or_indexes = list(set([str(r.features_names) for r in self]))
            else:
                fnames_or_indexes = list(set([str(r.features_indexes) for r in self]))
            dict_names = {}
            lmax = 1
            for f in fnames_or_indexes:
                l_ = len(ast.literal_eval(f))
                if l_ > lmax:
                    lmax = l_
                if l_ not in dict_names:
                    dict_names[l_] = []
                dict_names[l_].append(f)
            for l_ in dict_names:
                dict_names[l_].sort(reverse=reverse)
            fnames_or_indexes = []
            for l_ in range(1, lmax + 1):
                if l_ in dict_names:
                    fnames_or_indexes += dict_names[l_]

            rules_by_fnames = OrderedDict({f: [] for f in fnames_or_indexes})
            for rule in self:
                # noinspection PyUnresolvedReferences
                if which == "name":
                    v = str(rule.features_names)
                else:
                    v = str(rule.features_indexes)
                rules_by_fnames[v].append(rule)
            rules_by_fnames = {
                n: sorted(rules_by_fnames[n], key=lambda x: x.condition.bmins + x.condition.bmaxs)
                for n in rules_by_fnames
            }
            self._rules = []
            for n in rules_by_fnames:
                self._rules += rules_by_fnames[n]
        elif hasattr(self[0], criterion):
            self._rules = sorted(self, key=lambda x: getattr(x, criterion), reverse=reverse)
        else:
            raise ValueError(f"Can not sort RuleSet according to criterion {criterion}")
        if self.stack_activation:
            self.stacked_activations = self.stacked_activations[[str(r.condition) for r in self]]

    def compute_self_activation(self):
        """Computes the activation vector of self from its rules, using time-efficient Activation.multi_logical_or."""
        if len(self) == 0:
            return
        activations_available = all([r.activation_available for r in self])
        if activations_available:
            # noinspection PyProtectedMember
            self._activation = Activation.multi_logical_or([r._activation for r in self])

    def compute_stacked_activation(self):
        """Computes the stacked activation vectors of self from its rules."""
        if len(self) == 0:
            return
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("RuleSet's stacked activations requies pandas. Please run\npip install pandas")

        activations_available = all([r.activation_available for r in self])
        if activations_available:
            # noinspection PyProtectedMember
            self.stacked_activations = pd.DataFrame(
                data=np.array([r.activation for r in self]).T, columns=[str(r.condition) for r in self]
            )

    def del_activations(self):
        """Deletes the data, but not the relevent attributes, of the activation vector or each rules in self."""
        for r in self:
            r.del_activation()

    def del_activation(self):
        """Deletes the activation vector's data of self, but not the object itself, so any computed attribute remains
        available"""
        if hasattr(self, "_activation") and self._activation is not None:
            self._activation.delete()

    def del_stacked_activations(self):
        """Deletes stacked activation vectors of self. Set it to None."""
        if hasattr(self, "stacked_activations") and self.stacked_activations is not None:
            del self.stacked_activations
            self.stacked_activations = None

    def evaluate(self, xs: Union["pd.DataFrame", np.ndarray]) -> Activation:
        """Computes and returns the activation vector from an array of features.

        Parameters
        ----------
        xs: Union[pd.DataFrame, np.ndarray]
            The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
            or pd.DataFrame.

        Returns
        -------
        Activation
        """
        if len(self) == 0:
            raise ValueError("Can not use evaluate : The ruleset is empty!")
        activations = [rule.evaluate(xs) for rule in self.rules]
        return Activation.multi_logical_or(activations)

    def calc_activation(self, xs: Union[np.ndarray, "pd.DataFrame"]):
        """Uses input xs features data to compute the activation vector of all rules in self, and updates self's
        activation if self.remember_activation is True and stacked activation if self.stack_activation is True"""
        if len(self) == 0:
            raise ValueError("Can not use calc_activation : The ruleset is empty!")
        [rule.calc_activation(xs) for rule in self.rules]

        if self.remember_activation:
            self._activation = None
            self.compute_self_activation()
        if self.stack_activation:
            self.stacked_activations = None
            self.compute_stacked_activation()

    def get_features_count(self) -> List[Tuple[Any, int]]:
        """
        Get a counter of all different features in the ruleset. If names are not available, will use indexes.

        Returns:
        --------
        count : List[Tuple[Any, int]]
            Counter of all different features in the ruleset
        """
        # noinspection PyUnresolvedReferences
        if len(self) == 0:
            return []
        if len(self.features_names) > 0:
            var_in = list(itertools.chain(*[rule.features_names for rule in self]))
        else:
            var_in = list(itertools.chain(*[rule.feautres_indexes for rule in self]))
        count = Counter(var_in)

        count = count.most_common()
        return count

    def load(self, path, **kwargs):
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("'pandas' is needed to read a RuleSet from a file. Please run\npip install pandas")

        if hasattr(path, "read"):
            rules = path.read(**kwargs)
        else:
            rules = pd.read_csv(path, **kwargs)
        if rules.empty:
            self._rules = []
        else:
            self._rules = [self.series_to_rule(rules.loc[r]) for r in rules.index]

        if len(self._rules) > 0:
            self.rule_type = type(self._rules[0])
        if self.remember_activation:
            self.compute_self_activation()
        if self.stack_activation:
            self.compute_stacked_activation()
        self.features_names = list(set(traverse([rule.features_names for rule in self])))

    def save(self, path):

        try:
            import pandas as pd
        except ImportError:
            raise ImportError("'pandas' is needed to save a RuleSet to a file. Please run\npip install pandas")

        if len(self) == 0:
            if hasattr(path, "write"):
                path.write(pd.DataFrame())
            else:
                pd.DataFrame().to_csv(path)
            return

        idx = copy(RuleSet.condition_index)
        if self.rule_type == ClassificationRule:
            idx += RuleSet.classification_rule_index
        elif self.rule_type == RegressionRule:
            idx += RuleSet.regression_rule_index
        elif self.rule_type == Rule:
            idx += RuleSet.rule_index
        else:
            raise TypeError(f"Impossible to save rule type {self.rule_type}")

        dfs = [
            self.rule_to_series(
                (i, r),
                index=idx,
            )
            for i, r in enumerate(self.rules)
        ]
        if len(dfs) > 0:
            df = pd.concat(dfs, axis=1).T
        else:
            df = pd.DataFrame(columns=idx)
        if hasattr(path, "write"):
            path.write(df)
        else:
            df.to_csv(path)

    # noinspection PyProtectedMember
    @staticmethod
    def rule_to_series(irule: Tuple[int, Rule], index: list) -> "pd.Series":
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("'pandas' is needed to read a RuleSet from a file. Please run\npip install pandas")
        i = irule[0]
        rule = irule[1]
        name = f"R_{i}({len(rule)})"
        sr = pd.Series(data=[str(getattr(rule, ind)) for ind in index], name=name, index=index, dtype=str)
        return sr

    @staticmethod
    def series_to_rule(srule: "pd.Series") -> Rule:

        condition_index = {c: None for c in RuleSet.condition_index}

        for ind in srule.index:
            if ind == "Unnamed: 0":
                continue
            if ind not in RuleSet.classification_rule_index + RuleSet.condition_index:
                raise IndexError(f"Invalid rule attribute '{ind}'")

        if "std" in srule.index:
            rule = RegressionRule()
            rule_idx = copy(RuleSet.regression_rule_index)
        elif "criterion" in srule.index:
            rule = ClassificationRule()
            rule_idx = copy(RuleSet.classification_rule_index)
        else:
            rule = Rule()
            rule_idx = copy(RuleSet.rule_index)

        for rule_ind in srule.index:
            str_value = str(srule[rule_ind])
            if rule_ind in condition_index:
                condition_index[rule_ind] = ast.literal_eval(str_value)
                # setattr(condition, rule_ind, ast.literal_eval(str_value))
            elif rule_ind in rule_idx:
                if rule_ind == "activation":
                    setattr(rule, f"_{rule_ind}", Activation(str_value))
                elif str_value == "nan":
                    setattr(rule, f"_{rule_ind}", np.nan)
                elif rule_ind == "sign":
                    setattr(rule, f"_{rule_ind}", str_value)
                else:
                    if rule_ind == "prediction":  # Prediction can be a str in case of classification
                        try:
                            setattr(rule, f"_{rule_ind}", ast.literal_eval(str_value))
                        except ValueError:
                            setattr(rule, f"_{rule_ind}", str_value)
                    else:
                        setattr(rule, f"_{rule_ind}", ast.literal_eval(str_value))
            else:
                continue

        rule._condition = HyperrectangleCondition(**condition_index)
        return rule


def traverse(o, tree_types=(list, tuple, RuleSet)):
    """Yields each elementary elements from nested list or tuple

    Example
    -------
    >>> list(traverse([[[1, 2], 3,], [4, 5]]))
    [1, 2, 3, 4, 5]
    """
    if isinstance(o, tree_types):
        for value in o:
            for subvalue in traverse(value, tree_types):
                yield subvalue
    else:
        yield o



import numpy as np
import pandas as pd
from ruleskit import HyperrectangleCondition
import pytest

from ruleskit.condition import DuplicatedFeatures


@pytest.mark.parametrize(
    "fi, fn, bi, ba, impossible, err",
    [
        ([0, 1], None, [0, 1], [2, 3], False, None),
        ([0, 1], ["A", "B"], [0, 1], [2, 3], False, None),
        (None, ["A", "B"], [0, 1], [2, 3], False, None),
        (None, None, [0, 1], [2, 3], False, ValueError),
        ([0, 1], None, [0, 1, 2], [2, 3], False, ValueError),
        ([0, 1], None, [0, 1], [1, 2, 3], False, ValueError),
        ([0, 1], ["A", "B", "C"], [1, 2], [2, 3], False, ValueError),
        ([0.2, 1], ["A", "B"], [1, 2], [2, 3], False, TypeError),
        ([0, 1], [1, "B"], [0, 1], [2, 3], False, TypeError),
        ([0, 1], ["A", "B"], ["0", 1], [2, 3], False, TypeError),
        ([0, 1], ["A", "B"], [0, 1], [2, "3"], False, TypeError),
        ([0, 1], ["A", "B"], [2, 1], [0, 3], True, None),
        ([0, 0], ["A", "B"], [2, 1], [0, 3], True, DuplicatedFeatures),
        ([0, 1], ["A", "A"], [2, 1], [0, 3], True, DuplicatedFeatures),
    ],
)
def test_init(fi, fn, bi, ba, impossible, err):
    if err is not None:
        with pytest.raises(err):
            _ = HyperrectangleCondition(features_indexes=fi, features_names=fn, bmins=bi, bmaxs=ba)
    else:
        cond = HyperrectangleCondition(features_indexes=fi, features_names=fn, bmins=bi, bmaxs=ba)
        assert cond.features_indexes == fi if fi is not None else list(range(len(fn)))
        assert cond.features_names == fn if fn is not None else [f"X_{i}" for i in cond.features_indexes]
        assert cond.bmins == bi
        assert cond.bmaxs == ba
        assert cond.impossible is impossible


@pytest.mark.parametrize(
    "fi, fn, bi, ba, fie, fne, bie, bae, according_to",
    [
        ([0, 1], None, [0, 1], [2, 3], [0, 1], ["X_0", "X_1"], [0, 1], [2, 3], "index"),
        ([1, 0], None, [1, 0], [3, 2], [0, 1], ["X_0", "X_1"], [0, 1], [2, 3], "index"),
        ([0, 1], None, [0, 1], [2, 3], [0, 1], ["X_0", "X_1"], [0, 1], [2, 3], "name"),
        ([1, 0], None, [1, 0], [3, 2], [0, 1], ["X_0", "X_1"], [0, 1], [2, 3], "name"),
        ([1, 0], ["A", "B"], [1, 0], [3, 2], [1, 0], ["A", "B"], [1, 0], [3, 2], "name"),
        ([1, 0], ["A", "B"], [1, 0], [3, 2], [0, 1], ["B", "A"], [0, 1], [2, 3], "index"),
    ],
)
def test_sort(fi, fn, bi, ba, fie, fne, bie, bae, according_to):
    HyperrectangleCondition.SORT_ACCORDING_TO = according_to
    cond = HyperrectangleCondition(features_indexes=fi, features_names=fn, bmins=bi, bmaxs=ba)
    assert cond.features_indexes == fie
    assert cond.features_names == fne
    assert cond.bmins == bie
    assert cond.bmaxs == bae
    HyperrectangleCondition.SORT_ACCORDING_TO = "index"


@pytest.mark.parametrize(
    "x, condition, output, error",
    [
        (
            np.array([[1, 3], [3, 4], [2, np.nan]]),
            HyperrectangleCondition([0], bmins=[1], bmaxs=[2]),
            np.array([1, 0, 1]),
            None,
        ),
        (
            np.array([[1, 3], [3, 4], [2, np.nan]]),
            HyperrectangleCondition([1], bmins=[3], bmaxs=[5]),
            np.array([1, 1, 0]),
            None,
        ),
        (
            np.array([[1, 3], [3, 4], [2, np.nan]]),
            HyperrectangleCondition([0, 1], bmins=[1, 3], bmaxs=[2, 5]),
            np.array([1, 0, 0]),
            None,
        ),
        (
            np.array([[1, 3], [3, 4], [2, np.nan]]),
            HyperrectangleCondition([0, 3], bmins=[1, 3], bmaxs=[2, 5]),
            np.array([1, 0, 0]),
            IndexError,
        ),
        (
            pd.DataFrame(data=[[1, 3], [3, 4], [2, np.nan]], columns=["A", "B"]),
            HyperrectangleCondition([0, 3], features_names=["A", "B"], bmins=[1, 3], bmaxs=[2, 5]),
            np.array([1, 0, 0]),
            None,
        ),
        (
            pd.DataFrame(data=[[1, 3], [3, 4], [2, np.nan]], columns=["A", "B"]),
            HyperrectangleCondition([0, 3], features_names=["A", "C"], bmins=[1, 3], bmaxs=[2, 5]),
            np.array([1, 0, 0]),
            IndexError,
        ),
        (
            [[1, 3], [3, 4], [2, np.nan]],
            HyperrectangleCondition([0, 3], features_names=["A", "C"], bmins=[1, 3], bmaxs=[2, 5]),
            np.array([1, 0, 0]),
            AttributeError,
        ),
    ],
)
def test_evaluate(x, condition, output, error):
    if error is not None:
        with pytest.raises(error):
            _ = condition.evaluate(x)
    else:
        res = condition.evaluate(x)
        # noinspection PyUnresolvedReferences
        np.testing.assert_equal(res, output)


@pytest.mark.parametrize(
    "condition1, condition2, output",
    [
        (
            HyperrectangleCondition([0], bmins=[1], bmaxs=[3], features_names=["b"]),
            HyperrectangleCondition([1], bmins=[1], bmaxs=[2], features_names=["a"]),
            HyperrectangleCondition([0, 1], bmins=[1, 1], bmaxs=[2, 3], features_names=["a", "b"]),
        ),
        (
            HyperrectangleCondition([0], bmins=[1], bmaxs=[3], features_names=["b"]),
            HyperrectangleCondition([0], bmins=[1], bmaxs=[2], features_names=["a"]),
            "Some features with different names had same index in both conditions in __and__",
        ),
        (
            HyperrectangleCondition([0, 1], bmins=[0, 0], bmaxs=[10, 2], features_names=["a", "b"]),
            HyperrectangleCondition([0, 2], bmins=[10, 1], bmaxs=[20, 1], features_names=["a", "c"]),
            HyperrectangleCondition([0, 1, 2], bmins=[10, 0, 1], bmaxs=[10, 2, 1], features_names=["a", "b", "c"]),
        ),
        (
            HyperrectangleCondition([0, 1], bmins=[0, 0], bmaxs=[10, 2], features_names=["a", "b"]),
            HyperrectangleCondition([0, 1], bmins=[20, 1], bmaxs=[30, 1], features_names=["a", "c"]),
            "Some features with different names had same index in both conditions in __and__",
        ),
        (
            HyperrectangleCondition([0, 1], bmins=[0, 0], bmaxs=[10, 2], features_names=["a", "b"]),
            HyperrectangleCondition([1, 2], bmins=[20, 1], bmaxs=[30, 1], features_names=["a", "c"]),
            "Some features present in both conditions in __and__ have different indexes",
        ),
        (
            HyperrectangleCondition([0, 1], bmins=[0, 0], bmaxs=[10, 2], features_names=["a", "b"]),
            HyperrectangleCondition([0, 2], bmins=[20, 1], bmaxs=[30, 1], features_names=["a", "c"]),
            HyperrectangleCondition([0, 1, 2], bmins=[20, 0, 1], bmaxs=[10, 2, 1], features_names=["a", "b", "c"]),
        ),
    ],
)
def test_and(condition1, condition2, output):

    if isinstance(output, str):
        with pytest.raises(IndexError) as e:
            _ = condition1 & condition2
        assert output in str(e.value)
    else:
        res = condition1 & condition2
        np.testing.assert_equal(res, output)


@pytest.mark.parametrize(
    "attr, value, impossible_or_raise",
    [
        ("bmins", [0, 1], False),
        ("bmaxs", [0, 1], False),
        ("features_indexes", [0, 2], False),
        ("features_names", ["C", "D"], False),
        ("bmins", [0, 1, 2], f"Condition has 2 features but you gave 3 bmins"),
        ("bmaxs", [0, 1, 2], f"Condition has 2 features but you gave 3 bmaxs"),
        ("features_indexes", [0, 1, 2], f"Condition has 2 features but you gave 3 indexes"),
        ("features_names", ["C", "D", "E"], f"Condition has 2 features but you gave 3 names"),
        ("bmins", [2, 1], True),
        ("bmaxs", [1, 0], True),
    ],
)
def test_set_attr(attr, value, impossible_or_raise):
    cond = HyperrectangleCondition(features_indexes=[0, 1], features_names=["A", "B"], bmins=[0, 1], bmaxs=[1, 2])
    if isinstance(impossible_or_raise, str):
        with pytest.raises(IndexError) as e:
            setattr(cond, attr, value)
        assert impossible_or_raise in str(e.value)
    else:
        setattr(cond, attr, value)
        assert getattr(cond, attr) == value
        assert cond.impossible is impossible_or_raise



from typing import List, Union
import copy
import numpy as np
from ..rule import RegressionRule, ClassificationRule
from ..ruleset import RuleSet
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from ..condition import HyperrectangleCondition

# noinspection PyProtectedMember
from sklearn.tree import _tree
import logging

logger = logging.getLogger(__name__)


def extract_rules_from_tree(
    tree: Union[DecisionTreeClassifier, DecisionTreeRegressor],
    xmins: Union[List[float], np.ndarray],
    xmaxs: Union[List[float], np.ndarray],
    features_names: List[str] = None,
    classes_names: List[str] = None,
    get_leaf: bool = False,
    remember_activation: bool = True,
    stack_activation: bool = False,
) -> RuleSet:
    """To extract rules from a sklearn decision tree

    features are the list of X names.
    In the tree ('decision_tree' later in the code), the list of relevant Xs are represented
    by their position in the list of Xs that was fed to the tree at learning.
    So to access the name of the feature relevant for a given node in the tree,
    one has first to do decision_tree.feature[node] to access the feature index in x_names,
    then do x_names[decision_tree.feature[node]]

    xmin and xmax are Series.
    Index is a X name, values are the min and the man value of the X for a given Y.
    The Y was specified before calling this function so it does not appear here.

    If all_rules_list is not None, it will be appended with the extracted rules,
    and nothing is returned. Else, returns the extracted rules

    Later in the code:
    bmins is, for a given condition, the minimum value that X must have to be selected
    By analogy, bmaxs is the max value.

    Hence bmaxs must always be >= to bmins, and bmins and bmaxs must be inside the interval
    [xmin, xmax].

    When selecting Ys based on conditions, the criterion will be X > bmins and X <= bmaxs.
    bmins is excluded and bmaxs included. So when creating bmins, if it is equal to bmins,
    change it to be a bit less than xmins so that xmins will pass the criterion.

    In the tree, to each node corresponds a maximum value for the associated X (feature).
    This is accessed by decision_tree.threshold[node] and will be the bmaxs of one condition of a
    rule.

    It will later be the bmins of one condition of the "opposite" rule. Indeed if there
    is a condition saying something like 'If X < 3 then do that', there is also a
    condition saying 'if X >= 3 then do this'.

    At the end there will be 2 rules per node. So not only the leaves give a rule, but
    each branching in the tree.
    Rules do not contain any value for Ys. Just conditions on Xs. RICE will set the
    values in calc_stat.

    If print_visitor is specified, will return a string describing what happened step
    by step

    Parameters
    ----------
    tree : Union[sklearn.tree.DecisionTreeRegressor, sklearn.tree.DecisionTreeClassifier]
    features_names: List[str],
        the list of X names
    classes_names: List[str],
        the list of Y classes names
    xmins: Union[List[int], List[float], np.ndarray]
        min values of each xs, one entry per x
    xmaxs: Union[List[int], List[float], np.ndarray]
        max values of each xs, one entry per x
    get_leaf: Boolean type
        To return only the leaf of the tree.
    remember_activation: bool
        The activation of the RuleSet is the logical OR of the activation of all its rules. It is only computed
        if remember_activation is True. (default value = True)
    stack_activation: bool
        If True, the RuleSet will keep in memory 2-D np.ndarray containing the activations of all its rules. This
        can take a lot of memory, but can save time if you apply numpy methods on this stacked vector instead of on
        each rule separately. (default value = False)

    Returns
    -------
    rule_list: List[rule.Rule]
        The extracted rules for mthe tree.

    """
    decision_tree = tree.tree_
    if features_names is None:
        features_names = ["X_" + str(i) for i in range(tree.n_features_)]

    def visitor(node, depth, condition=None, rules_list=None):
        if rules_list is None:
            rules_list = RuleSet(remember_activation=remember_activation, stack_activation=stack_activation)
        if decision_tree.feature[node] != _tree.TREE_UNDEFINED:
            # If
            new_condition = HyperrectangleCondition(
                features_indexes=[decision_tree.feature[node]],
                bmins=[xmins[decision_tree.feature[node]]],
                bmaxs=[decision_tree.threshold[node]],
                features_names=[features_names[decision_tree.feature[node]]],
            )

            # If cond is not None, means we are not at the first node,
            # so need to expand the list of conditions, unless the current X is
            # already in our list of conditions. In that case, update the
            # corresponding condition by setting its bmaxs to be the min of the
            # current and new thresh.olds
            if condition is not None:
                if decision_tree.feature[node] not in condition.features_indexes:
                    # Will concatenate current condition list and new conditions by their
                    # attributes (which are lists)
                    new_condition = condition & new_condition
                else:
                    new_bmaxs = decision_tree.threshold[node]
                    new_condition = copy.deepcopy(condition)
                    place = condition.features_indexes.index(decision_tree.feature[node])
                    new_condition.bmaxs[place] = min(new_bmaxs, new_condition.bmaxs[place])

            # Create a new Rule with all the condition and append it to the rules
            # list. So the rule list is actually a history of how one rule evolved.
            if get_leaf is False:
                if isinstance(tree, DecisionTreeClassifier):
                    new_rule = ClassificationRule(copy.deepcopy(new_condition))
                    if classes_names is None:
                        new_rule._prediction = [pred for pred in enumerate(decision_tree.value[node][0])]
                    else:
                        new_rule._prediction = [
                            (classes_names[i], pred) for i, pred in enumerate(decision_tree.value[node][0])
                        ]
                else:
                    new_rule = RegressionRule(copy.deepcopy(new_condition))
                    new_rule._prediction = decision_tree.value[node][0][0]
                rules_list.append(new_rule)

            # Execute the current function on the left of the current node
            # (the "True" side)
            rules_list = visitor(decision_tree.children_left[node], depth + 1, new_condition, rules_list)

            # At this point, any rule found on the left of the current node will be
            # in rules list and new_cond will contain the corresponding conditions.

            # Create a new condition corresponding to the opposite of the current
            # node's threshold, i.e bmins is now the previous bmaxs
            # Here we do not have to alter bmins, because we explicitly want
            # node_thresh to not pass the criterion since it must have passed it
            # for the opposite condition.
            new_condition = HyperrectangleCondition(
                features_indexes=[decision_tree.feature[node]],
                bmins=[decision_tree.threshold[node]],
                bmaxs=[xmaxs[decision_tree.feature[node]]],
                features_names=[features_names[decision_tree.feature[node]]],
            )

            # Means we are not at first node. Same logic as previously, except this time
            # if Xi is in the list of features then the new bmins is now the threshold
            # and bmaxs is modified to be the X maximum. Then keep the max of current
            # and new bmins as bmins, same for bmaxs.
            # TODO there is something weird here, for X_i will necessarily be in cond, since
            # it was added at the very beginning of the function.
            if condition is not None:
                if decision_tree.feature[node] not in condition.features_indexes:
                    new_condition = condition & new_condition
                else:
                    new_bmins = decision_tree.threshold[node]
                    new_bmaxs = xmaxs[decision_tree.feature[node]]
                    new_condition = copy.deepcopy(condition)
                    place = new_condition.features_indexes.index(decision_tree.feature[node])
                    new_condition.bmins[place] = max(new_bmins, new_condition.bmins[place])
                    new_condition.bmaxs[place] = max(new_bmaxs, new_condition.bmaxs[place])

            # Create the rule for the right side of the node then apply this
            # function on the right side of the node
            if get_leaf is False:
                if isinstance(tree, DecisionTreeClassifier):
                    new_rule = ClassificationRule(copy.deepcopy(new_condition))
                    if classes_names is None:
                        new_rule._prediction = [pred for pred in enumerate(decision_tree.value[node][0])]
                    else:
                        new_rule._prediction = [
                            (classes_names[i], pred) for i, pred in enumerate(decision_tree.value[node][0])
                        ]
                else:
                    new_rule = RegressionRule(copy.deepcopy(new_condition))
                    new_rule._prediction = decision_tree.value[node][0][0]
                rules_list.append(new_rule)

            rules_list = visitor(decision_tree.children_right[node], depth + 1, new_condition, rules_list)

        elif get_leaf:
            if isinstance(tree, DecisionTreeClassifier):
                new_rule = ClassificationRule(copy.deepcopy(condition))
                if classes_names is None:
                    new_rule._prediction = [pred for pred in enumerate(decision_tree.value[node][0])]
                else:
                    new_rule._prediction = [
                        (classes_names[i], pred) for i, pred in enumerate(decision_tree.value[node][0])
                    ]
                rules_list.append(new_rule)
            else:
                new_rule = RegressionRule(copy.deepcopy(condition))
                new_rule._prediction = decision_tree.value[node][0][0]
                rules_list.append(new_rule)

        return rules_list

    rule_list = visitor(0, 1)
    return rule_list



from ruleskit import Activation, Rule, RegressionRule, ClassificationRule
import pytest


@pytest.fixture
def clean():
    yield
    Activation.clean_files()


@pytest.fixture
def prepare_fit():
    calc_attributes_r = Rule.calc_attributes
    calc_attributes_rr = RegressionRule.calc_attributes
    calc_attributes_cr = ClassificationRule.calc_attributes
    yield
    Rule.calc_attributes = calc_attributes_r
    RegressionRule.calc_attributes = calc_attributes_rr
    ClassificationRule.calc_attributes = calc_attributes_cr



import math
import logging
from typing import Union
from pathlib import Path
from json import load

logger = logging.getLogger(__name__)


class Thresholds:

    """Implements the notion of thresholds in Rule. From a given json, sets limits that can be used to flag bad rules.

    Attributes
    ----------
    limits: dict
        The limits to use
    path: Union[str, Path, TransparentPath]
    """

    KWOWN_ARGS = {"abs": abs, "sqrt": math.sqrt, "exp": math.exp, "log": math.log, "str": str, "int": int}

    def __init__(self, path: Union[str, Path, "TransparentPath"], show=False):
        self.limits = {}
        self.path = path

        if type(path) == str:
            try:
                from transparentpath import TransparentPath
                path = TransparentPath(path)
            except ImportError:
                path = Path(path)

        if not path.is_file():
            return

        if show:
            logger.info(f"Found threshold file {path}")

        if hasattr(path, "read"):
            self.limits = path.read()
        else:
            with open(path) as opath:
                self.limits = load(opath)

        if show:
            message = "\n".join([f"{i}: {self.limits[i]}" for i in self.limits])
            logger.info(f"Thresholds are \n{message}")
        if "coverage" not in self.limits:
            self.limits["coverage"] = {"min": 0.05}
            logger.info("Coverage limit was not set. Set to minimum 5%.")
        elif "min" not in self.limits["coverage"]:
            self.limits["coverage"]["min"] = 0.05
            logger.info("Coverage lower limit was not set. Set to 5%.")
        elif self.limits["coverage"]["min"] == 0:
            logger.warning("Coverage limit is set to 0.")

    def __call__(self, tkey: str, rule) -> bool:
        if tkey not in self.limits:
            return True
        value = getattr(rule, tkey)
        if callable(value):
            return True
        if value is None:
            return True

        threshold = self.limits[tkey]
        s = tkey
        if "arg" in threshold:
            if threshold["arg"] not in Thresholds.KWOWN_ARGS:
                raise ValueError(f"Unknown arg for threshold : {threshold['arg']}")
            value = Thresholds.KWOWN_ARGS[threshold["arg"]](value)
            s = f"{threshold['arg']}({tkey})"

        for lim in threshold:
            if lim == "arg":
                continue
            if isinstance(threshold[lim], str):
                threshold[lim] = getattr(rule, threshold[lim])
            if not isinstance(threshold[lim], (int, float)):
                raise TypeError(f"Value in threshols can only be float or int, got {type(threshold[lim])}")

        if "min" in threshold and value < threshold["min"]:
            logger.debug(f"Rule {rule} is bad : {s} = {value} < {threshold['min']}")
            return False
        if "max" in threshold and value > threshold["max"]:
            logger.debug(f"Rule {rule} is bad : {s} = {value} > {threshold['max']}")
            return False
        if "equal" in threshold and value != threshold["equal"]:
            logger.debug(f"Rule {rule} is bad : {s} = {value} != {threshold['equal']}")
            return False
        if "different" in threshold and value == threshold["different"]:
            logger.debug(f"Rule {rule} is bad : {s} = {value} == {threshold['different']}")
            return False
        return True



from ruleskit import Thresholds


class Dummy:
    def __init__(self, **kwargs):
        self.crit_max = 2
        for key in kwargs:
            setattr(self, key, kwargs[key])


def test_thresholds():
    ts = Thresholds("tests/unit/data/thresholds.json")
    dum = Dummy(coverage=0.03)
    assert not ts("coverage", dum)
    assert ts("coverage", Dummy(coverage=0.06))
    assert ts("zscore", Dummy(zscore=2))
    assert ts("zscore", Dummy(zscore=-2))
    assert not ts("zscore", Dummy(zscore=1.5))
    assert not ts("zscore", Dummy(zscore=-1.5))
    assert not ts("crit_mean", Dummy(crit_mean=-1.5))
    assert ts("crit_mean", Dummy(crit_mean=1.5))
    assert ts("crit_min", Dummy(crit_min=1))
    assert ts("crit_min", Dummy(crit_min=-1))
    assert not ts("crit_min", Dummy(crit_min=2.1))
    assert not ts("crit_min", Dummy(crit_min=-2.1))


from ruleskit import Rule, RegressionRule, ClassificationRule, HyperrectangleCondition, Activation
from bitarray import bitarray
from pathlib import Path
import numpy as np
import pandas as pd
import pytest
import itertools


# noinspection PyUnresolvedReferences,PyCallingNonCallable
@pytest.mark.parametrize(
    "condition_activation_error, theclass",
    itertools.product(
        [
            (
                HyperrectangleCondition(features_indexes=[0, 1], features_names=["A", "B"], bmins=[0, 1], bmaxs=[2, 3]),
                Activation(np.array([0, 1, 1, 0]), to_file=False),
                None,
            ),
            (None, None, None),
            (None, Activation(np.array([0, 1, 1, 0]), to_file=False), ValueError),
            (
                HyperrectangleCondition(features_indexes=[0, 1], features_names=["A", "B"], bmins=[0, 1], bmaxs=[2, 3]),
                None,
                None,
            ),
        ],
        [Rule, RegressionRule, ClassificationRule],
    ),
)
def test_init(clean, condition_activation_error, theclass):
    condition, activation, error = condition_activation_error
    if error is not None:
        with pytest.raises(error):
            _ = theclass(condition, activation)
    else:
        r = theclass(condition, activation)
        assert r.condition == condition
        if activation is not None:
            np.testing.assert_equal(r.activation, activation.raw)
        else:
            assert r.activation is None
        assert r._activation == activation


@pytest.mark.parametrize(
    "x_y_condition_activation_s_error, theclass",
    itertools.product(
        [
            (
                np.array([[1, 3], [3, 4], [2, np.nan]]),
                np.array([1, 3, 2]),
                HyperrectangleCondition([0], bmins=[1], bmaxs=[2]),
                np.array([1, 0, 1]),
                "101",
                None,
            ),
            (
                np.array([[1, 3], [3, 4], [2, np.nan]]),
                np.array([1, 3, 2]),
                HyperrectangleCondition([1], bmins=[3], bmaxs=[5]),
                np.array([1, 1, 0]),
                "110",
                None,
            ),
            (
                np.array([[1, 3], [3, 4], [2, np.nan]]),
                np.array([1, 3, 2]),
                HyperrectangleCondition([0, 1], bmins=[1, 3], bmaxs=[2, 5]),
                np.array([1, 0, 0]),
                "100",
                None,
            ),
            (
                pd.DataFrame(data=[[1, 3], [3, 4], [2, np.nan]], columns=["A", "B"]),
                np.array([1, 3, 2]),
                HyperrectangleCondition([0, 1], features_names=["A", "B"], bmins=[1, 3], bmaxs=[2, 5]),
                np.array([1, 0, 0]),
                "100",
                None,
            ),
            (
                pd.DataFrame(data=[[1, 3], [3, 4], [2, np.nan]], columns=["A", "B"]),
                np.array([1, 3, 2]),
                HyperrectangleCondition([0, 1], features_names=["A", "B"], bmins=[1, 3], bmaxs=[2, 5]),
                None,
                "100",
                ValueError,
            ),
        ],
        [Rule, RegressionRule, ClassificationRule],
    ),
)
def test_fit(clean, prepare_fit, x_y_condition_activation_s_error, theclass):
    x, y, condition, activation, s, error = x_y_condition_activation_s_error
    # noinspection PyCallingNonCallable
    rule = theclass(condition=condition)
    if error is not None:

        def wrong_calc_attributes(r, y_, **kwargs):
            return

        theclass.calc_attributes = wrong_calc_attributes
        # noinspection PyTypeChecker
        with pytest.raises(error) as e:
            rule.fit(xs=x, y=y)
        assert "'fit' did not set 'prediction'" in str(e.value)
    elif theclass == Rule:
        with pytest.raises(NotImplementedError):
            rule.fit(xs=x, y=y)
    else:
        rule.fit(xs=x, y=y)
        np.testing.assert_equal(rule.activation, activation)
        np.testing.assert_equal(rule._activation.as_bitarray, bitarray(s))
        assert rule._activation.as_integer == int(s, 2)


@pytest.mark.parametrize(
    "x, y, condition, activation",
    [
        (
            np.array([[1, 3], [3, 4], [2, np.nan]]),
            np.array([1, 3, 2]),
            HyperrectangleCondition([0, 1], bmins=[1, 3], bmaxs=[2, 5]),
            np.array([1, 0, 0]),
        ),
    ],
)
def test_local_activation(clean, x, y, condition, activation):
    Rule.LOCAL_ACTIVATION = False
    rule = RegressionRule(condition=condition)
    rule.fit(xs=x, y=y)
    assert not isinstance(rule._activation.data, Path)
    Rule.LOCAL_ACTIVATION = True
    rule = RegressionRule(condition=condition)
    rule.fit(xs=x, y=y)
    assert isinstance(rule._activation.data, Path) and rule._activation.data.is_file()


@pytest.mark.parametrize(
    "x, y, condition, cov, pred",
    [
        (
            np.array([[1, 3], [3, 4], [2, np.nan]]),
            np.array([1, 3, 2]),
            HyperrectangleCondition([0], bmins=[1], bmaxs=[2]),
            2 / 3,
            1.5,
        ),
        (
            np.array([[1, 3], [3, 4], [2, np.nan]]),
            np.array([1, 3, 2]),
            HyperrectangleCondition([1], bmins=[3], bmaxs=[5]),
            2 / 3,
            2,
        ),
        (
            np.array([[1, 3], [3, 4], [2, np.nan]]),
            np.array([1, 3, 2]),
            HyperrectangleCondition([0, 1], bmins=[1, 3], bmaxs=[2, 5]),
            1 / 3,
            1,
        ),
    ],
)
def test_regression_attributes(clean, x, y, condition, cov, pred):
    rule = RegressionRule(condition=condition)
    rule.fit(xs=x, y=y)
    np.testing.assert_equal(rule.coverage, cov)
    np.testing.assert_equal(rule.prediction, pred)


@pytest.mark.parametrize(
    "x, y, condition, proba, pred, crit",
    [
        (
            np.array([[1, 3], [3, 4], [2, np.nan]]),
            np.array(["a", "b", "a"]),
            HyperrectangleCondition([0], bmins=[1], bmaxs=[2]),
            [("a", 1.0)],
            "a",
            1.0,
        ),
        (
            np.array([[1, 3], [3, 4], [2, np.nan]]),
            np.array(["a", "b", "a"]),
            HyperrectangleCondition([1], bmins=[4], bmaxs=[5]),
            [("b", 1.0)],
            "b",
            1.0,
        ),
        (
            np.array([[1, 3], [2, 4], [5, np.nan]]),
            np.array(["a", "b", "a"]),
            HyperrectangleCondition([0], bmins=[1], bmaxs=[5]),
            [("a", 2 / 3), ("b", 1 / 3)],
            "a",
            2 / 3,
        ),
    ],
)
def test_classification_attributes(clean, x, y, condition, proba, pred, crit):
    rule = ClassificationRule(condition=condition)
    rule.fit(xs=x, y=y)
    np.testing.assert_equal(rule._prediction, proba)
    np.testing.assert_equal(rule.prediction, pred)
    np.testing.assert_equal(rule.criterion, crit)


@pytest.mark.parametrize(
    "x, y, condition1, condition2, activation1, activation2, activation_test",
    [
        (
            np.array([[1, 3], [3, 4], [2, np.nan]]),
            np.array([1, 3, 2]),
            HyperrectangleCondition([0], bmins=[1], bmaxs=[2]),
            HyperrectangleCondition([1], bmins=[3], bmaxs=[5]),
            np.array([1, 0, 1]),
            np.array([1, 1, 0]),
            np.array([1, 0, 0]),
        ),
    ],
)
def test_and(clean, x, y, condition1, condition2, activation1, activation2, activation_test):
    rule1 = RegressionRule(condition=condition1)
    rule1.fit(xs=x, y=y)
    rule2 = RegressionRule(condition=condition2)
    rule2.fit(xs=x, y=y)

    new_rule = rule1 & rule2
    np.testing.assert_equal(new_rule.activation, activation_test)
    new_rule.fit(xs=x, y=y)
    np.testing.assert_equal(new_rule.activation, activation_test)


@pytest.mark.parametrize(
    "condition_x_y_x2_expected, theclass",
    itertools.product(
        [
            (
                HyperrectangleCondition([0, 1], bmins=[1, 3], bmaxs=[2, 5]),
                np.array([[1, 3], [3, 4], [2, np.nan]]),
                np.array([1.5, 3, 2]),
                np.array([[1, 3], [0, 4], [0, 5], [np.nan, np.nan]]),
                np.array([1.5, np.nan, np.nan, np.nan]),
            ),
            (
                HyperrectangleCondition([0, 1], bmins=[1, 3], bmaxs=[2, 5]),
                np.array([[1, 3], [3, 4], [2, np.nan]]),
                np.array([1.5, 3, 2]),
                None,
                np.array([1.5, np.nan, np.nan]),
            ),
            (
                HyperrectangleCondition([0, 1], features_names=["A", "B"], bmins=[1, 3], bmaxs=[2, 5]),
                pd.DataFrame(data=[[1, 3], [3, 4], [2, np.nan]], columns=["A", "B"]),
                np.array([1.5, 3, 2]),
                np.array([[1, 3], [0, 4], [0, 5], [np.nan, np.nan]]),
                np.array([1.5, np.nan, np.nan, np.nan]),
            ),
            (
                HyperrectangleCondition([0, 1], features_names=["A", "B"], bmins=[1, 3], bmaxs=[2, 5]),
                pd.DataFrame(data=[[1, 3], [3, 4], [2, np.nan]], columns=["A", "B"]),
                np.array([1.5, 3, 2]),
                None,
                np.array([1.5, np.nan, np.nan]),
            ),
            (
                HyperrectangleCondition([0, 1], features_names=["A", "B"], bmins=[1, 3], bmaxs=[2, 5]),
                np.array([[1, 3], [3, 4], [2, np.nan]]),
                np.array([1.5, 3, 2]),
                pd.DataFrame([[1, 3], [0, 4], [0, 5], [np.nan, np.nan]], columns=["A", "B"]),
                np.array([1.5, np.nan, np.nan, np.nan]),
            ),
            (
                HyperrectangleCondition([0, 1], features_names=["A", "B"], bmins=[1, 3], bmaxs=[2, 5]),
                pd.DataFrame(data=[[1, 3], [3, 4], [2, np.nan]], columns=["A", "B"]),
                np.array([1.5, 3, 2]),
                pd.DataFrame([[1, 3], [0, 4], [0, 5], [np.nan, np.nan]], columns=["A", "B"]),
                np.array([1.5, np.nan, np.nan, np.nan]),
            ),
            (
                HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                np.array([[3, 1], [4, 3], [np.nan, 2]]),
                np.array([1.5, 3, 2]),
                np.array([[3, 1], [4, 0], [5, 0], [np.nan, np.nan]]),
                np.array([1.5, np.nan, np.nan, np.nan]),
            ),
            (
                HyperrectangleCondition([1, 0], features_names=["A", "B"], bmins=[1, 3], bmaxs=[2, 5]),
                np.array([[3, 1], [4, 3], [np.nan, 2]]),
                np.array([1.5, 3, 2]),
                np.array([[3, 1], [4, 0], [5, 0], [np.nan, np.nan]]),
                np.array([1.5, np.nan, np.nan, np.nan]),
            ),
            (
                HyperrectangleCondition([0, 1], features_names=["B", "A"], bmins=[1, 3], bmaxs=[2, 5]),
                pd.DataFrame([[3, 1], [4, 3], [np.nan, 2]], columns=["A", "B"]),
                np.array([1.5, 3, 2]),
                pd.DataFrame([[3, 1], [4, 0], [5, 0], [np.nan, np.nan]], columns=["A", "B"]),
                np.array([1.5, np.nan, np.nan, np.nan]),
            ),
            (
                HyperrectangleCondition([0, 1], features_names=["B", "A"], bmins=[1, 3], bmaxs=[2, 5]),
                pd.DataFrame([[3, 1], [4, 3], [np.nan, 2]], columns=["A", "B"]),
                np.array(["a", "b", "c"]),
                pd.DataFrame([[3, 1], [4, 0], [5, 0], [np.nan, np.nan]], columns=["A", "B"]),
                np.array(["a", "nan", "nan", "nan"]),
            ),
        ],
        [Rule, RegressionRule, ClassificationRule],
    ),
)
def test_predict(clean, condition_x_y_x2_expected, theclass):
    condition, x, y, x2, expected = condition_x_y_x2_expected
    # noinspection PyCallingNonCallable
    rule = theclass(condition=condition)
    if y.dtype == str and theclass != ClassificationRule:
        return

    if theclass == Rule:
        with pytest.raises(NotImplementedError):
            rule.fit(xs=x, y=y)
        return
    if y.dtype.type == np.str_ and theclass != ClassificationRule:
        return

    rule.fit(xs=x, y=y)
    pred = rule.predict(x2)
    if theclass == ClassificationRule:
        np.testing.assert_equal(pred, expected)
    else:
        np.testing.assert_almost_equal(pred, expected)


@pytest.mark.parametrize(
    "rule1, rule2, pred1, pred2, expected",
    [
        (
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 1]), to_file=False)),
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 1, 0]), to_file=False)),
            0.2,
            0.3,
            1/3,
        ),
        (
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 1]), to_file=False)),
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 1, 0]), to_file=False)),
            0.2,
            -0.3,
            -1/3,
        ),
        (
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 1]), to_file=False)),
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 1, 0]), to_file=False)),
            -0.2,
            -0.3,
            1/3,
        ),
        (
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 1]), to_file=False)),
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 0]), to_file=False)),
            0.2,
            0.3,
            2/3,
        ),
        (
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 1]), to_file=False)),
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 0]), to_file=False)),
            -0.2,
            0.3,
            -2/3,
        ),
        (
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 1]), to_file=False)),
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 0]), to_file=False)),
            -0.2,
            -0.3,
            2/3,
        ),
        (
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 1]), to_file=False)),
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 1]), to_file=False)),
            0.2,
            0.3,
            1.0,
        ),
        (
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([1, 0, 1]), to_file=False)),
            Rule(condition=HyperrectangleCondition([1, 0], bmins=[1, 3], bmaxs=[2, 5]),
                 activation=Activation(np.array([0, 1, 0]), to_file=False)),
            0.2,
            0.3,
            0.0,
        ),
    ],
)
def test_correlation(clean, rule1, rule2, pred1, pred2, expected):
    rule1._prediction = pred1
    rule2._prediction = pred2
    assert rule1.get_correlation(rule2) == expected



import itertools

import numpy as np
from ruleskit import Activation
import pytest
import pandas as pd
from pathlib import Path
from bitarray import bitarray

data_path = Path("tests/unit/data")


def make_column(will_compare, optimize, store_raw, to_file, ignore) -> str:
    s = []
    if ignore is None:
        return f"will_compare_{will_compare}_optimize_{optimize}_store_raw_{store_raw}_to_file_{to_file}"
    if "will_compare" not in ignore:
        s.append(f"will_compare_{will_compare}")
    if "optimize" not in ignore:
        s.append(f"optimize_{optimize}")
    if "store_raw" not in ignore:
        s.append(f"store_raw_{store_raw}")
    if "to_file" not in ignore:
        s.append(f"to_file_{to_file}")
    return "_".join(s)


def compare_column(res, df, column):

    print(column)
    for index in df[column].index:
        print(index)
        attr = getattr(res, index)
        expected_attr = df.loc[index, column]
        if index == "data_format":
            assert attr == expected_attr
        elif expected_attr == "=0":
            assert attr == 0
        elif expected_attr == ">0":
            assert attr > 0
        elif expected_attr == "-1":
            assert attr == -1
        elif expected_attr == "None":
            assert attr is None
        elif expected_attr == "not None":
            assert attr is not None
        elif expected_attr == "osef":
            continue
        else:
            assert attr == int(expected_attr)


# noinspection PyTypeChecker
@pytest.mark.parametrize(
    "vector_cs_ca_b_i_n_p_c_o_no, will_compare, optimize, store_raw, to_file, withwhat",
    itertools.product(
        [
            (
                    np.array(
                        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]),
                    "1,1,275,276",
                    np.array([1, 1, 275, 276]),
                    bitarray(
                        "1000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
                        "0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
                        "0000000000000000000000000000000000000000000000000000000000000000000000000000000000000001"),
                    60708402882054033466233184588234965832575213720379360039119137804340758912662765569,
                    276,
                    data_path / "a1.txt",
                    2 / 276,
                    [0, 275],
                    2,
            ),
            (
                np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1]),
                "0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16",
                np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]),
                bitarray("0101010101010101"),
                21845,
                16,
                data_path / "a2.txt",
                0.5,
                [1, 3, 5, 7, 9, 11, 13, 15],
                8,
            ),
        ],
        [True, False],
        [True, False],
        [True, False],
        [True, False],
        ["raw", "integer", "bitarray", "compressed_str", "compressed_array", "file"],
    )
)
def test_init(clean, vector_cs_ca_b_i_n_p_c_o_no, will_compare, optimize, store_raw, to_file, withwhat):
    vector, cs, ca, b, i, n, p, c, o, no = vector_cs_ca_b_i_n_p_c_o_no
    print(withwhat, p)
    Activation.WILL_COMPARE = will_compare
    Activation.STORE_RAW = store_raw
    value = vector
    to_file = False
    if withwhat == "integer":
        value = i
        with pytest.raises(ValueError):
            _ = Activation(value, optimize=optimize, to_file=to_file)
    if withwhat == "bitarray":
        value = b
    if withwhat == "compressed_str":
        value = cs
    if withwhat == "compressed_array":
        value = ca
    if withwhat == "file":
        value = p

    expected_df = pd.read_csv(data_path / f"{p.stem}_activation_init_{withwhat}.csv", index_col=0)
    ignore = expected_df.index.name

    column = make_column(will_compare, optimize, store_raw, to_file, ignore)

    res = Activation(value, optimize=optimize, to_file=to_file, length=n)
    Activation.WILL_COMPARE = False
    Activation.STORE_RAW = False

    compare_column(res, expected_df, column)

    np.testing.assert_equal(res.raw, vector)
    np.testing.assert_equal(res.as_bitarray, b)
    np.testing.assert_equal(res.as_integer, i)
    np.testing.assert_equal(res.as_compressed_str, cs)
    np.testing.assert_equal(res.as_compressed_array, ca)
    np.testing.assert_equal(res.length, n)
    assert res.entropy == len(ca)
    assert res.rel_entropy == len(ca) / n
    assert res.coverage == c
    assert res.ones == o
    assert res.nones == no
    if to_file:
        assert res.data.is_file()

    expected_df = pd.read_csv(data_path / f"{p.stem}_activation_init_{withwhat}_after_calls.csv", index_col=0)
    ignore = expected_df.index.name

    column = make_column(will_compare, optimize, store_raw, to_file, ignore)

    compare_column(res, expected_df, column)


@pytest.mark.parametrize(
    "vector1, vector2, diff",
    [
        (
            np.array([1, 0, 1]),
            np.array([1, 1, 0]),
            np.array([0, 0, 1]),
        ),
    ],
)
def test_diff(clean, vector1, vector2, diff):
    act1 = Activation(vector1)
    act2 = Activation(vector2)
    np.testing.assert_equal((act1 - act2).raw, diff)


@pytest.mark.parametrize(
    "vector1, vector2, and_vector",
    [
        (
            np.array([1, 0, 1]),
            np.array([1, 1, 0]),
            np.array([1, 0, 0]),
        ),
    ],
)
def test_and(clean, vector1, vector2, and_vector):
    act1 = Activation(vector1)
    act2 = Activation(vector2)
    np.testing.assert_equal((act1 & act2).raw, and_vector)


@pytest.mark.parametrize(
    "vector1, vector2, add_vector",
    [
        (
            np.array([1, 0, 1]),
            np.array([1, 1, 0]),
            np.array([1, 1, 1]),
        ),
    ],
)
def test_or(clean, vector1, vector2, add_vector):
    act1 = Activation(vector1)
    act2 = Activation(vector2)
    comb = (act1 | act2).raw
    np.testing.assert_equal(comb, add_vector)


@pytest.mark.parametrize(
    "vector1, vector2, add_vector",
    [
        (
            np.array([1, 0, 1]),
            np.array([1, 1, 0]),
            np.array([0, 1, 1]),
        ),
    ],
)
def test_xor(clean, vector1, vector2, add_vector):
    act1 = Activation(vector1)
    act2 = Activation(vector2)
    comb = (act1 ^ act2).raw
    np.testing.assert_equal(comb, add_vector)


@pytest.mark.parametrize(
    "vector1, vector2, add_vector",
    [
        (
            np.array([1, 0, 1]),
            np.array([1, 1, 0]),
            np.array([0, 0, 1]),
        ),
    ],
)
def test_sub(clean, vector1, vector2, add_vector):
    act1 = Activation(vector1)
    act2 = Activation(vector2)
    comb = (act1 - act2).raw
    np.testing.assert_equal(comb, add_vector)


@pytest.mark.parametrize(
    "vector1, vector2, expected",
    [
        (
            np.array([1, 0, 1]),
            np.array([1, 1, 0]),
            1/3,
        ),
        (
            np.array([1, 0, 1]),
            np.array([1, 0, 0]),
            2/3,
        ),
        (
            np.array([1, 0, 1]),
            np.array([1, 0, 1]),
            1.0,
        ),
        (
            np.array([1, 0, 1]),
            np.array([0, 1, 0]),
            0.0,
        ),
    ],
)
def test_correlation(clean, vector1, vector2, expected):
    assert Activation(vector1, to_file=False).get_correlation(Activation(vector2, to_file=False)) == expected



from sklearn.tree import DecisionTreeRegressor
from sklearn import datasets
from ruleskit import extract_rules_from_tree, HyperrectangleCondition, Rule, RuleSet
import numpy as np
import pytest


@pytest.mark.parametrize(
    "tree, output, get_leaf",
    [
        (
            DecisionTreeRegressor(max_depth=2),
            RuleSet([
                Rule(
                    HyperrectangleCondition(
                        [1], [-0.126097385560409], [-0.0037617861526086926]
                    )
                ),
                Rule(
                    HyperrectangleCondition(
                        [1, 0],
                        [-0.126097385560409, -0.0902752958985185],
                        [-0.0037617861526086926, 0.0061888848431408405],
                    )
                ),
                Rule(
                    HyperrectangleCondition(
                        [1, 0],
                        [-0.126097385560409, 0.0061888848431408405],
                        [-0.0037617861526086926, 0.17055522598066],
                    )
                ),
                Rule(
                    HyperrectangleCondition(
                        [1], [-0.0037617861526086926], [0.133598980013008]
                    )
                ),
                Rule(
                    HyperrectangleCondition(
                        [1, 0],
                        [-0.0037617861526086926, -0.0902752958985185],
                        [0.133598980013008, 0.014811381697654724],
                    )
                ),
                Rule(
                    HyperrectangleCondition(
                        [1, 0],
                        [-0.0037617861526086926, 0.014811381697654724],
                        [0.133598980013008, 0.17055522598066],
                    )
                ),
            ]),
            False,
        ),
        (
            DecisionTreeRegressor(max_depth=2),
            RuleSet([
                Rule(
                    HyperrectangleCondition(
                        [1, 0],
                        [-0.126097385560409, -0.0902752958985185],
                        [-0.0037617861526086926, 0.0061888848431408405],
                    )
                ),
                Rule(
                    HyperrectangleCondition(
                        [1, 0],
                        [-0.126097385560409, 0.0061888848431408405],
                        [-0.0037617861526086926, 0.17055522598066],
                    )
                ),
                Rule(
                    HyperrectangleCondition(
                        [1, 0],
                        [-0.0037617861526086926, -0.0902752958985185],
                        [0.133598980013008, 0.014811381697654724],
                    )
                ),
                Rule(
                    HyperrectangleCondition(
                        [1, 0],
                        [-0.0037617861526086926, 0.014811381697654724],
                        [0.133598980013008, 0.17055522598066],
                    )
                ),
            ]),
            True,
        ),
    ],
)
def test_extract_rules_from_tree(tree, output, get_leaf):
    diabetes = datasets.load_diabetes()
    data = diabetes.data
    # Keep the two most important variables
    x = data.T[2]
    y = data.T[-2]
    x_vstack = np.vstack((x, y)).T
    # Get the target variable
    y = diabetes.target

    tree.fit(x_vstack, y)
    rule_list = extract_rules_from_tree(
        tree, xmins=x_vstack.min(axis=0), xmaxs=x_vstack.max(axis=0), get_leaf=get_leaf
    )
    np.testing.assert_equal(rule_list, output)


# @pytest.mark.parametrize(
#     "tree",
#     [
#             DecisionTreeRegressor(max_depth=2),
#     ],
# )
# def test_extract_rules_from_tree(tree):
#     diabetes = datasets.load_diabetes()
#     data = diabetes.data
#     # Keep the two most important variables
#     x = data.T[2]
#     y = data.T[-2]
#     X = np.vstack((x, y)).T
#     # Get the target variable
#     y = diabetes.target
#
#     tree.fit(X, y)
#     prediction_tree = tree.predict(X)
#
#     rule_list = extract_rules_from_tree(tree, xmins=X.min(axis=0), xmaxs=X.max(axis=0))
#     [rule.fit(X, y) for rule in rule_list]
#     prediction_rules = np.zeros(len(y))
#     for rule in rule_list:
#         prediction_rules += rule.predict(X)
#
#     np.testing.assert_equal(prediction_rules, prediction_tree)



import numpy as np
from collections import Counter
from typing import Union, Tuple, List
import logging

logger = logging.getLogger(__name__)


def most_common_class(
    activation: Union[np.ndarray, None], y: np.ndarray
) -> List[Tuple[str, float]]:
    if activation is None:
        return np.bincount(y).argmax()

    if isinstance(activation, np.ndarray):
        y_conditional = np.extract(activation, y)
    else:
        raise TypeError("'activation' in conditional_mean must be None or a np.ndarray")
    count = Counter(y_conditional)
    n = len(y_conditional)
    prop = [v / n for v in count.values()]
    return [(c, v) for c, v in zip(count.keys(), prop)]


def conditional_mean(activation: Union[np.ndarray, None], y: np.ndarray) -> float:
    """Mean of all activated values

    If activation is None, we assume the given y have already been extracted from the activation vector,
    which saves time.
    """
    if activation is None:
        return float(np.nanmean(y))

    if isinstance(activation, np.ndarray):
        y_conditional = np.extract(activation, y)
    else:
        raise TypeError("'activation' in conditional_mean must be None or a np.ndarray")
    non_nans_conditional_y = y_conditional[~np.isnan(y_conditional)]
    if len(non_nans_conditional_y) == 0:
        logger.debug("None of the activated points have a non-nan value in target y. Conditional mean is set to 0.")
        return 0
    return float(np.mean(non_nans_conditional_y))


def conditional_std(activation: Union[np.ndarray, None], y: np.ndarray) -> float:
    """Standard deviation of all activated values

    If activation is None, we assume the given y have already been extracted from the activation vector,
    which saves time.
    """
    if activation is None:
        return float(np.nanstd(y))

    if isinstance(activation, np.ndarray):
        y_conditional = np.extract(activation, y)
    else:
        raise TypeError("'activationt' in conditional_std must be None or a np.ndarray")
    return float(np.nanstd(y_conditional))


def mse_function(prediction_vector: np.ndarray, y: np.ndarray) -> float:
    """
    Compute the mean squared error
    "$ \\dfrac{1}{n} \\Sigma_{i=1}^{n} (\\hat{y}_i - y_i)^2 $"

    Parameters
    ----------
    prediction_vector : np.ndarray
        A predictor vector. It means a sparse array with two
        different values ymean, if the rule is not active
        and the prediction is the rule is active.

    y : np.ndarray
        The real target values (real numbers)

    Return
    ------
    criterion : float
        the mean squared error
    """
    if len(prediction_vector) != len(y):
        raise ValueError("The two array must have the same length")
    error_vector = prediction_vector - y
    criterion = np.nanmean(error_vector ** 2)
    # noinspection PyTypeChecker
    return criterion


def mae_function(prediction_vector: np.ndarray, y: np.ndarray) -> float:
    """
    Compute the mean absolute error
    "$ \\dfrac{1}{n} \\Sigma_{i=1}^{n} |\\hat{y}_i - y_i| $"

    Parameters
    ----------
    prediction_vector : np.ndarray
        A predictor vector. It means a sparse array with two
        different values ymean, if the rule is not active
        and the prediction is the rule is active.

    y : np.ndarray
        The real target values (real numbers)

    Return
    ------
    criterion : float
        the mean absolute error
    """
    if len(prediction_vector) != len(y):
        raise ValueError("The two array must have the same length")
    error_vect = np.abs(prediction_vector - y)
    criterion = np.nanmean(error_vect)
    # noinspection PyTypeChecker
    return criterion


def aae_function(prediction_vector: np.ndarray, y: np.ndarray) -> float:
    """
    Compute the mean squared error
    "$ \\dfrac{1}{n} \\Sigma_{i=1}^{n} (\\hat{y}_i - y_i)$"

    Parameters
    ----------
    prediction_vector : np.ndarray
        A predictor vector. It means a sparse array with two
        different values ymean, if the rule is not active
        and the prediction is the rule is active.

    y : np.ndarray
        The real target values (real numbers)

    Return
    ------
    criterion : float
        the mean squared error
    """
    if len(prediction_vector) != len(y):
        raise ValueError("The two array must have the same length")
    error_vector = np.mean(np.abs(prediction_vector - y))
    median_error = np.mean(np.abs(y - np.median(y)))
    return error_vector / median_error


def calc_regression_criterion(prediction_vector: np.ndarray, y: np.ndarray, **kwargs) -> float:
    """
    Compute the criterion

    Parameters
    ----------
    prediction_vector : np.ndarray
        The prediction vector

    y : np.ndarray
        The real target values (real numbers)

    kwargs : dict
        Can contain 'method', the method mse_function or mse_function criterion (default is 'mse'), and 'cond', whether
         to evaluate the criterion only if the rule is activated (default is True)

    Return
    ------
    criterion : float
        Criteria value
    """

    method = kwargs.get("method", "mse")
    cond = kwargs.get("cond", True)

    if cond:
        sub_y = np.extract(prediction_vector != 0, y)
        sub_pred = np.extract(prediction_vector != 0, prediction_vector)
    else:
        sub_y = y
        sub_pred = prediction_vector

    if method.lower() == "mse":
        criterion = mse_function(sub_pred, sub_y)

    elif method.lower() == "mae":
        criterion = mae_function(sub_pred, sub_y)

    elif method.lower() == "aae":
        criterion = aae_function(sub_pred, sub_y)

    else:
        raise ValueError(f"Unknown criterion: {method}. Please choose among mse, mae and aae")

    return criterion


def success_rate(prediction: Union[int, str], y: np.ndarray):
    success = sum(y == prediction)
    return success / len(y)


# noinspection PyUnresolvedReferences
def calc_classification_criterion(
    activation_vector: np.ndarray, prediction: Union[int, str], y: Union[np.ndarray, "pd.Series"], **kwargs
) -> float:
    """
    Computes the criterion

    Parameters
    ----------
    activation_vector : Union[np.ndarray, pd.Series]
        The prediction vector

    prediction: int or str:
                The label prediction

    y : np.ndarray
        The real target values (real numbers)

    kwargs : dict
        Can contain 'method', the method mse_function or mse_function criterion (default is 'mse'), and 'cond', whether
         to evaluate the criterion only if the rule is activated (default is True)

    Return
    ------
    criterion : float
        Criteria value
    """

    method = kwargs.get("method", "success_rate")
    cond = kwargs.get("cond", True)

    if cond:
        sub_y = np.extract(activation_vector != 0, y)
    else:
        sub_y = y

    if method.lower() == "success_rate":
        criterion = success_rate(prediction, sub_y)

    else:
        raise ValueError(f"Unknown criterion: {method}. Please choose among success_rate")

    return criterion



import ast
from abc import ABC
from copy import copy
from typing import List, Union, Tuple
import numpy as np
from numbers import Number
from .activation import Activation
import logging

logger = logging.getLogger(__name__)


class DuplicatedFeatures(Exception):
    pass


# noinspection PyUnresolvedReferences
class Condition(ABC):
    """Abstract class for Condition object. Used by Rule objets.
    A condition is a list of variable (here represented by their indexes in an array) and of conditions on those
    variables.
    A condition can be "imossible" to meet, in that case self.impossible is True. This is set automatically.

    One can add conditions and use logical AND (&) operations on two conditions (same thing as add). In that case, the
    two conditions are combined into a new one.
    """

    def __init__(self, features_indexes: Union[List[int], None] = None, empty: bool = False):
        if empty:
            self._features_indexes = None
        else:
            if features_indexes is None:
                raise ValueError("Must specify features_indexes")
            if any([not np.issubdtype(type(a), np.integer) for a in features_indexes]):
                raise TypeError(
                    f"features_indexes must be integers. You gave {[(f, type(f)) for f in features_indexes]}"
                )
            self._features_indexes = features_indexes
            if len(set(self._features_indexes)) != len(self._features_indexes):
                raise DuplicatedFeatures

        self.impossible = False

    def sort(self):
        self._features_indexes = sorted(self._features_indexes)

    def __and__(self, other: "Condition") -> "Condition":
        """To be implemented in daughter classes"""
        pass

    def __add__(self, other: "Condition") -> "Condition":
        return NotImplemented("Can not add conditions (seen as 'logical OR'). You can use logical AND however.")

    @property
    def to_hash(self):
        return ("c",) + tuple(self._features_indexes)

    def __hash__(self):
        return hash(frozenset(self.to_hash))

    @property
    def features_indexes(self) -> List[int]:
        return self._features_indexes

    @features_indexes.setter
    def features_indexes(self, value: Union[List[int], str]):
        if isinstance(value, str):
            value = ast.literal_eval(value)
        if len(set(value)) != len(value):
            raise DuplicatedFeatures
        if len(self) > 0 and len(self) != len(value):
            raise IndexError(f"Condition has {len(self)} features but you gave {len(value)} indexes")
        self._features_indexes = value

    def __len__(self):
        """A Condition's length is the number of features it talks about"""
        return len(self._features_indexes)

    def evaluate(self, xs: Union["pd.DataFrame", np.ndarray]) -> np.ndarray:
        """
        To be implemented in daughter class.
        Evaluates where a condition if fullfilled. In this abstract class that does not have any acutal condition,
        it is always fullfilled.

        Parameters
        ----------
        xs: Union[pd.DataFrame, np.ndarray]
            shape (n, d), n number of line, d number of features

        Returns
        -------
        activation: Activation
             Shape  (n, 1). The activation vector, filled with 1 where the condition is met and 0 where it is not.
        """
        pass

    def normalize_features_indexes(self):
        """In some daughter classes, features indexes are optional. however since the attribute 'features_indexes'
        must be specified, a default value is set automatically"""
        self._features_indexes = list(range(len(self.features_indexes)))


# noinspection PyUnresolvedReferences
class HyperrectangleCondition(Condition):
    """Condition class for Hyper Rectangle conditions.

    An Hyper Rectangle condition is a condition where each feature is associated to a min and a max (self.bmins and
    self.bmaxs). The condition is met when all features are within their respective bmin and bmax.

    For example, if the condition has :

    features_indexes = [0, 1]
    bminx = [0, 1]
    mnaxs = [0, 2]

    Then the condition is met when feature 0 is equal to 0 and feature 2 is between 1 and 2.

    In this condition, features indexes are optional if features names are given.

    Such a condition can be sorted either according to features indexes (smaller features first) or by features names in
    alphabetical order. This is set through the class attribute SORT_ACCORDING_TO that can be either "index" or "name".
    """

    SORT_ACCORDING_TO = "index"

    def __init__(
            self,
            features_indexes: Union[List[int], None] = None,
            bmins: Union[List[Union[int, float]], None] = None,
            bmaxs: Union[List[Union[int, float]], None] = None,
            features_names: Union[List[str], None] = None,
            empty: bool = False,
            sort: bool = True,
    ):
        if empty:
            super().__init__(empty=True)
            self._bmins = None
            self._bmaxs = None
            self._features_names = None
        else:
            if bmins is None:
                raise ValueError("bmins can not be None if 'empty' is False")
            if bmaxs is None:
                raise ValueError("bmaxs can not be None if 'empty' is False")
            if features_indexes is None:
                if features_names is None:
                    raise ValueError("Must specify at least one of features_indexes and features_names")
                features_indexes = list(range(len(features_names)))

            length = len(features_indexes)
            if len(bmaxs) != length:
                raise ValueError(f"Specifed {length} features but {len(bmaxs)} bmaxs")
            if len(bmins) != length:
                raise ValueError(f"Specifed {length} features but {len(bmins)} bmins")
            if features_names is not None and len(features_names) != length:
                raise ValueError(f"Specifed {length} features but {len(features_names)} bmaxs")

            if features_names is not None and any([not isinstance(a, str) for a in features_names]):
                raise TypeError(f"Names must be strings. You gave {[(f, type(f)) for f in features_names]}")
            if any([not isinstance(a, Number) for a in bmins]):
                raise TypeError(f"bmins must be integers or floats. You gave {[(f, type(f)) for f in bmins]}")
            if any([not isinstance(a, Number) for a in bmaxs]):
                raise TypeError(f"bmaxs must be integers or floats. You gave {[(f, type(f)) for f in bmaxs]}")

            super().__init__(features_indexes)

            if any([a > b for a, b in zip(bmins, bmaxs)]):
                # If a bmin is above its associated bmax, then the rule is impossible.
                self.impossible = True
            self._bmins = bmins
            self._bmaxs = bmaxs
            if features_names is not None:
                self._features_names = features_names
            else:
                self._features_names = ["X_" + str(i) for i in self._features_indexes]
            if len(set(self._features_names)) != len(self._features_names):
                raise DuplicatedFeatures
            if sort:
                self.sort()

    def __and__(self, other: "HyperrectangleCondition") -> "HyperrectangleCondition":

        """Logical and (&) or two HyperrectangleCondition objects

        If the two conditions do not talk about the same features, then the AND is obvious. For common features,
        the bmin of the feature in the new condition is set to be the greatest of bmins in parent conditions, and the
        bmax the smallest of bmaxs.
        This can give impossible conditions : if in condition 1 feature A must be between 0 and 10 and in condition 2 it
        must be between 20 and 30, then in the new condition it must be between 20 and 10, assumin 20 being the minimum.
        In that case, the new condition, upon creation, will have self.impossible = True. This does not corrupt the
        object nor the code : the condition's method "evaluate", which returns the activation vector, will return a
        vector with only zeros since the condition will never be met.
        """

        self_clone = HyperrectangleCondition(
            features_indexes=copy(self.features_indexes),
            bmins=copy(self.bmins),
            bmaxs=copy(self.bmaxs),
            features_names=copy(self.features_names),
        )

        common_features = [f for f in self_clone.features_names if f in other.features_names]

        if len(common_features) > 0:

            # If the two conditions have features in common, the new conditons will have as range the intersection of
            # each condition's range for those features. The new condition can possibly never be met.

            common_features_positions_in_self = [self_clone.features_names.index(f) for f in common_features]
            common_features_positions_in_other = [other.features_names.index(f) for f in common_features]

            (common_features_indexes_in_self, common_features_bmins_in_self, common_features_bmaxs_in_self) = list(zip(
                *[
                    (
                        self_clone.features_indexes[i],
                        self_clone.bmins[i],
                        self_clone.bmaxs[i]
                    )
                    for i in common_features_positions_in_self
                ]
            ))

            (common_features_indexes_in_other, common_features_bmins_in_other, common_features_bmaxs_in_other) = list(
                zip(
                    *[
                        (
                            other.features_indexes[i],
                            other.bmins[i],
                            other.bmaxs[i]
                        )
                        for i in common_features_positions_in_other
                    ]
                )
            )

            if common_features_indexes_in_self != common_features_indexes_in_other:
                raise IndexError("Some features present in both conditions in __and__ have different indexes : \n "
                                 f"{common_features_indexes_in_self}\n "
                                 f"{common_features_indexes_in_other}")

            common_features_bmins = [
                max(bmin0, bmin1) for bmin0, bmin1 in zip(common_features_bmins_in_self, common_features_bmins_in_other)
            ]
            common_features_bmaxs = [
                min(bmax0, bmax1) for bmax0, bmax1 in zip(common_features_bmaxs_in_self, common_features_bmaxs_in_other)
            ]

            features_indexes = [
                other.features_indexes[i]
                for i in range(len(other.features_indexes))
                if i not in common_features_positions_in_other
            ]
            features_names = [f for f in other.features_names if f not in common_features]
            bmins = [
                other.bmins[i]
                for i in range(len(other.bmins))
                if i not in common_features_positions_in_other
            ]
            bmaxs = [
                other.bmaxs[i]
                for i in range(len(other.bmaxs))
                if i not in common_features_positions_in_other
            ]

            other_clone = HyperrectangleCondition(
                features_indexes=features_indexes,
                bmins=bmins,
                bmaxs=bmaxs,
                features_names=features_names,
            )

            for i, index in enumerate(common_features_positions_in_self):
                self_clone.bmins[index] = common_features_bmins[i]
                self_clone.bmaxs[index] = common_features_bmaxs[i]
        else:
            other_clone = other

        args = [i + j for i, j in zip(self_clone.getattr, other_clone.getattr)]
        if len(set(args[0])) != len(args[0]):
            raise IndexError("Some features with different names had same index in both conditions in __and__:\n "
                             f"{args}")

        to_ret = HyperrectangleCondition(
            features_indexes=args[0],
            bmins=args[1],
            bmaxs=args[2],
            features_names=args[3],
            empty=False,
        )
        return to_ret

    @property
    def getattr(self) -> List[list]:
        return [self.features_indexes, self.bmins, self.bmaxs, self.features_names]

    @property
    def features_names(self) -> List[str]:
        return self._features_names

    @property
    def bmins(self) -> List[Union[int, float]]:
        return self._bmins

    @property
    def bmaxs(self) -> List[Union[int, float]]:
        return self._bmaxs

    @features_names.setter
    def features_names(self, value: Union[List[str], str]):
        if isinstance(value, str):
            value = ast.literal_eval(value)
        if len(set(value)) != len(value):
            raise DuplicatedFeatures
        if len(self) > 0 and len(self) != len(value):
            raise IndexError(f"Condition has {len(self)} features but you gave {len(value)} names")
        self._features_names = value

    @bmins.setter
    def bmins(self, value: Union[List[Union[int, float]], str]):
        if isinstance(value, str):
            value = [int(v) for v in ast.literal_eval(value)]
        if len(self) > 0 and len(self) != len(value):
            raise IndexError(f"Condition has {len(self)} features but you gave {len(value)} bmins")
        self._bmins = value
        if any([a > b for a, b in zip(self.bmins, self.bmaxs)]):
            # If a bmin is above its associated bmax, then the rule is impossible.
            self.impossible = True

    @bmaxs.setter
    def bmaxs(self, value: Union[List[Union[int, float]], str]):
        if isinstance(value, str):
            value = [int(v) for v in ast.literal_eval(value)]
        if len(self) > 0 and len(self) != len(value):
            raise IndexError(f"Condition has {len(self)} features but you gave {len(value)} bmaxs")
        self._bmaxs = value
        if any([a > b for a, b in zip(self.bmins, self.bmaxs)]):
            # If a bmin is above its associated bmax, then the rule is impossible.
            self.impossible = True

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if self._features_names is None:
            return "empty condition"
        str_output = f"{self._features_names[0]} in [{self._bmins[0]}, {self._bmaxs[0]}]"
        if len(self) > 1:
            for i in range(1, len(self)):
                str_output += " AND "
                str_output += f"{self._features_names[i]} in [{self._bmins[i]}, {self._bmaxs[i]}]"
        return str_output

    def __eq__(self, other):
        """Two HyperrectangleConditions are equal if they talk about the same features, and if the bmins and bmaxs
        are the same from one rule to another. Features indexes can be different."""
        return self.__hash__() == other.__hash__()

    @property
    def to_hash(self) -> Tuple[str]:
        return ("c",) + tuple(
            (self._features_names[i], self._bmins[i], self._bmaxs[i]) for i in range(len(self._features_names))
        )

    def __hash__(self):
        return hash(frozenset(self.to_hash))

    def __getitem__(self, item):
        return (
            self._features_names[item],
            self._features_indexes[item],
            self._bmins[item],
            self._bmaxs[item],
        )

    def __len__(self) -> int:
        """A HyperrectangleCondition's length is the number of features it talks about"""
        if self._features_names is not None:
            return len(self._features_names)
        if self._features_indexes is not None:
            return len(self._features_indexes)
        if self._bmins is not None:
            return len(self._bmins)
        if self._bmaxs is not None:
            return len(self._bmaxs)
        return 0

    def sort(self):
        if len(self) > 1:
            if HyperrectangleCondition.SORT_ACCORDING_TO == "index":
                self._bmins = [x for _, x in sorted(zip(self._features_indexes, self._bmins))]
                self._bmaxs = [x for _, x in sorted(zip(self._features_indexes, self._bmaxs))]
                self._features_names = [x for _, x in sorted(zip(self._features_indexes, self._features_names))]
                self._features_indexes = sorted(self._features_indexes)
            elif HyperrectangleCondition.SORT_ACCORDING_TO == "name":
                self._bmins = [x for _, x in sorted(zip(self._features_names, self._bmins))]
                self._bmaxs = [x for _, x in sorted(zip(self._features_names, self._bmaxs))]
                self._features_indexes = [x for _, x in sorted(zip(self._features_names, self._features_indexes))]
                self._features_names = sorted(self._features_names)
            else:
                raise ValueError(
                    "HyperrectangleCondition's SORT_ACCORDING_TO"
                    f" can be 'index' or 'name', not {HyperrectangleCondition.SORT_ACCORDING_TO}"
                )

    def evaluate(self, xs: Union["pd.DataFrame", np.ndarray]) -> np.ndarray:
        """
        Evaluates where a condition if fullfilled, by returning a vector of the form [0, 1, 0, 0, ...]

        Parameters
        ----------
        xs: Union[pd.DataFrame, np.ndarray]
            shape (n, d), n number of line, d number of features. If is a pd.DataFrame, will use self.features_names to
            select features to use in xs

        Returns
        -------
        activation: np.ndarray
            Shape  (n, 1). The activation vector, filled with 1 where the condition is met and 0 where it is not.

        Examples
        --------
        >>> xs_ = np.array([[1, 3], [3, 4], [2, np.nan]])
        >>> c1 = HyperrectangleCondition([0], bmins=[1], bmaxs=[2])
        >>> c1.evaluate(xs_)
        np.array([1, 0, 1])
        >>> c2 = HyperrectangleCondition([1], bmins=[3], bmaxs=[5])
        >>> c2.evaluate(xs_)
        np.array([1, 1, 0])
        >>> c3 = HyperrectangleCondition([0, 1], bmins=[1, 3], bmaxs=[2, 5])
        >>> c3.evaluate(xs_)
        np.array([1, 0, 0])
        """
        if self.impossible:
            return np.zeros(xs.shape[0], dtype=np.ubyte)

        if isinstance(xs, np.ndarray):
            if any([i >= xs.shape[1] for i in self.features_indexes]):
                raise IndexError("Some features indexes in self are greater than the size of the given xs array")
            geq_min = leq_min = not_nan = np.ones(xs.shape[0], dtype=np.ubyte)
            for i, j in enumerate(self._features_indexes):
                geq_min &= np.greater_equal(xs[:, j], self._bmins[i])
                leq_min &= np.less_equal(xs[:, j], self._bmaxs[i])
                not_nan &= np.isfinite(xs[:, j])
            activation = geq_min & leq_min & not_nan
        else:
            if any([i not in xs.columns for i in self.features_names]):
                raise IndexError("Some features names in self were not in xs DataFrame columns")
            geq_min = leq_min = not_nan = np.ones(xs.shape[0], dtype=np.ubyte)
            for i, n in enumerate(self._features_names):
                geq_min &= np.greater_equal(xs[n], self._bmins[i])
                leq_min &= np.less_equal(xs[n], self._bmaxs[i])
                not_nan &= np.isfinite(xs[n])
            activation = (geq_min & leq_min & not_nan).values

        return activation



from .activation import Activation
from .condition import Condition, HyperrectangleCondition, DuplicatedFeatures
from .rule import Rule, RegressionRule, ClassificationRule
from .ruleset import RuleSet
from .thresholds import Thresholds
from .utils.rule_utils import extract_rules_from_tree

try:
    from ._version import __version__
except ImportError:
    pass



from abc import ABC
import numpy as np
from typing import Optional, Union, Tuple
from time import time
from pathlib import Path
from .condition import Condition
from .activation import Activation
from .utils import rfunctions as functions
from .thresholds import Thresholds
import logging

logger = logging.getLogger(__name__)


# noinspection PyUnresolvedReferences
class Rule(ABC):

    """An abstract Rule object.

    A Rule is a condition (represented by any daughter class of ruleskit.Condition), applied on real features and target
    data.
    The Rule contains, in addition to the Condition object, many attributes dependent on the features data, such as
    the activation vector (a 1-D np.ndarray with 0 when the rule is activated - condition is met - and 0 when it is not)
    but also the rule's prediction (computed in the daughter class).

    Daughter classes can remember more attributes (precision, user-definded criterion...).

    Rule also include metrics that can be used for profiling the code : it will remember the time taken to fit the rule
    (fitting is the computation of the rule's attribute from the condition and the features data), the time taken
    to compute the activation vector and the time taken to make a prediction.

    To compute those metrics, one must use the rule's "fit" methods. Once this is done, one cas use the "predict"
     methods on a different set of features data.

    The Rule object can access any attribute of its condition as if it was its own : rule.features_indexes will return
    the features_indexes attribute's value of the condition in the Rule object. See Condition class for more details.

    The Rule object can also access any attribute of its activation vector as if it was its own. See Activation class
    for more details.
    """

    LOCAL_ACTIVATION = True
    THRESHOLDS = None
    """Thresholds that the Rule must meet to be good. See `ruleskit.thresholds.Thresholds` for more details."""

    @classmethod
    def SET_THRESHOLDS(cls, path: Union[str, Path, "TransparentPath", None], show=False):
        """Set thresholds globally for all futur Rules"""
        if path is None:
            cls.THRESHOLDS = None
        else:
            cls.THRESHOLDS = Thresholds(path, show)

    def __init__(
        self, condition: Optional[Condition] = None, activation: Optional[Activation] = None,
    ):

        if condition is not None and not isinstance(condition, Condition):
            raise TypeError("Argument 'condition' must derive from Condition or be None.")
        if activation is not None and not isinstance(activation, Activation):
            raise TypeError("Argument 'activation' must derive from Activation or be None.")
        if activation is not None and condition is None:
            raise ValueError("Condition can not be None if activation is not None")

        self._condition = condition
        self._activation = activation
        self._thresholds = Rule.THRESHOLDS
        if self._activation is not None:
            self.check_thresholds("coverage")

        self._coverage = None
        self._prediction = None

        self._time_fit = -1
        self._time_calc_activation = -1
        self._time_predict = -1

        self._good = True
        self._bad_because = None

    def set_thresholds(self, path: Union[str, Path, "TransparentPath"], show=False):
        """Set thresholds for this rule only"""
        if path is None:
            cls.THRESHOLDS = None
        else:
            self._thresholds = Thresholds(path, show)

    def check_thresholds(self, attribute: Optional[str] = None) -> None:
        """If `ruleskit.rule.Rule.THRESHOLDS` is specified, will check that this rule is good regarding those
        thresholds, and set the flags *good* and *bad_because* accordingly

        Parameters
        ----------
        attribute: Optional[str]
            If specified, will only check the threshold of this rule attribute. If not, will test every rule attributes
            for which a threshold is defined.
        """

        if Rule.THRESHOLDS is None:
            return

        if attribute is not None:
            if not Rule.THRESHOLDS(attribute, self):
                self._bad_because = attribute
                self._good = False
            return

        for attribute in dir(self):
            if attribute.startswith("__"):
                continue
            if not Rule.THRESHOLDS(attribute, self):
                self._bad_because = attribute
                self._good = False
                return
        logger.debug(f"Rule {self} is good")

    @property
    def coverage(self) -> float:
        if self._activation is not None:
            self._coverage = self._activation.coverage
            return self._activation.coverage
        return self._coverage

    @coverage.setter
    def coverage(self, value):
        if self._activation is not None:
            self._activation.coverage = value
        self._coverage = value

    def __and__(self, other: "Rule") -> "Rule":
        """Logical AND (&) of two rules. It is simply the logical AND of the two rule's conditions and activations. """
        condition = self._condition & other._condition
        activation = self._activation & other._activation
        return self.__class__(condition, activation)

    def __add__(self, other: "Rule") -> "Rule":
        return NotImplemented("Can not add rules (seen as 'logical OR'). You can use logical AND however.")

    # def __del__(self):
    #     self.del_activation()

    def del_activation(self):
        """Deletes the activation vector's data, but not the object itself, so any computed attributes will remain
        available"""
        if hasattr(self, "_activation") and self._activation is not None:
            self._activation.delete()

    @property
    def activation_available(self) -> bool:
        """Returns True if the rule has an activation vector, and if this Activation's object data is available."""
        if self._activation is None:
            return False
        if self._activation.data_format == "file":
            return self._activation.data.is_file()
        else:
            return self._activation.data is not None

    @property
    def condition(self) -> Condition:
        return self._condition

    @property
    def activation(self) -> Union[None, np.ndarray]:
        """Returns the Activation vector's data in a form of a 1-D np.ndarray, or None if not available.

        Returns
        -------
        np.ndarray
            of the form [0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, ...]
        """
        if self._activation:
            return self._activation.raw
        return None

    @property
    def prediction(self) -> Union[str, float]:
        return self._prediction

    @property
    def thresholds(self) -> Thresholds:
        return self._thresholds

    @property
    def good(self) -> bool:
        return self._good

    @property
    def bad_because(self) -> str:
        return self._bad_because

    @property
    def time_fit(self) -> float:
        """Profiling attribute. Time in seconds taken to fit the rule"""
        return self._time_fit

    @property
    def time_predict(self) -> float:
        """Profiling attribute. Time in seconds taken by the rule to make a prediction"""
        return self._time_predict

    @property
    def time_calc_activation(self) -> float:
        """Profiling attribute. Time in seconds taken to comptue the activation vector"""
        return self._time_calc_activation

    def __getattr__(self, item):
        """If item is not found in self, try to fetch it from its activation or condition."""
        if item == "_activation" or item == "_condition":
            raise AttributeError(f"'Rule' object has no attribute '{item}'.")

        if hasattr(self._activation, item):
            return getattr(self._activation, item)
        if hasattr(self._condition, item):
            return getattr(self._condition, item)
        raise AttributeError(f"'Rule' object has no attribute '{item}'.")

    def __setattr__(self, item, value):
        """If item is private (starts with _), then default behavior. Else, if the item is not yet known by the rule
        but is known by its condition or activation, will set it to the condition or the activation. Else,
        raises AttributeError."""
        if item.startswith("_"):
            super(Rule, self).__setattr__(item, value)
            return
        if not hasattr(self, item):
            if hasattr(self._activation, item):
                setattr(self._activation, item, value)
            elif hasattr(self._condition, item):
                setattr(self._condition, item, value)
            else:
                raise AttributeError(f"Can not set attribute '{item}' in object Rule.")
        else:
            super(Rule, self).__setattr__(item, value)

    def __eq__(self, other) -> bool:
        """Two rules are equal if their conditions are equal."""
        if not isinstance(other, Rule):
            return False
        else:
            return self._condition == other._condition

    def __contains__(self, other: "Rule") -> bool:
        """A Rule contains another Rule if the second rule's activated points are also all activated by the first rule.
        """
        if not self._activation or not other._activation:
            return False
        return other._activation in self._activation

    def __str__(self) -> str:
        prediction = "<prediction unset>"
        if self._prediction is not None:
            prediction = self._prediction
        if self._condition is None:
            return "empty rule"
        return f"If {self._condition.__str__()} Then {prediction}."

    @property
    def to_hash(self) -> Tuple[str]:
        return ("r",) + self._condition.to_hash[1:]

    def __hash__(self) -> hash:
        return hash(frozenset(self.to_hash))

    def __len__(self):
        """A Rule's length is the number of features it talks about"""
        return len(self._condition)

    def evaluate(self, xs: Union["pd.DataFrame", np.ndarray]) -> Activation:
        """Computes and returns the activation vector from an array of features.

        Parameters
        ----------
        xs: Union[pd:DataFrame, np.ndarray]
            The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
            or pd:DataFrame.

        Returns
        -------
        Activation
        """
        arr = self._condition.evaluate(xs)
        # noinspection PyTypeChecker
        a = Activation(arr, to_file=Rule.LOCAL_ACTIVATION)
        return a

    # noinspection PyUnusedLocal
    def fit(self, xs: Union["pd.DataFrame", np.ndarray] = None, y: np.ndarray = None, **kwargs):
        """Computes activation, and other criteria dependant on the nature of the daughter class of the Rule,
        for a given xs and y.

        Parameters
        ----------
        xs: Union[pd:DataFrame, np.ndarray]
            The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
            or pd:DataFrame.
        y: np.ndarray
            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray.
        kwargs: dict
            Other arguments used by daughter class
        """
        t0 = time()
        if xs is not None:
            self.calc_activation(xs)

        if self._activation is None:
            raise ValueError("If fitting without specifying xs, activation must have been computed already.")
        self.calc_attributes(y, **kwargs)
        if self.prediction is None:
            raise ValueError("'fit' did not set 'prediction' : did you overload 'calc_attributes' correctly ?")

        self._time_fit = time() - t0

    def calc_attributes(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
        """Implement in daughter class. Must set self._prediction."""
        raise NotImplementedError("To implement in daughter class")

    def calc_activation(self, xs: Union["pd.DataFrame", np.ndarray]):
        """Uses self.evaluate to set self._activation.

        Parameters
        ----------
        xs: Union[pd:DataFrame, np.ndarray]
            The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
            or pd:DataFrame.
        """
        t0 = time()
        self._activation = self.evaluate(xs)
        self._time_calc_activation = time() - t0
        self.check_thresholds("coverage")

    def predict(self, xs: Optional[Union["pd.DataFrame", np.ndarray]] = None) -> Union[np.ndarray, "pd.Series"]:
        """Returns the prediction vector. If xs is not given, will use existing activation vector.
        Will raise ValueError is xs is None and activation is not yet known.

        Parameters
        ----------
        xs: Optional[Union[pd:DataFrame, np.ndarray]]
            The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
            or pd:DataFrame. If not specified the rule's activation vector must have been computed already.

        Returns
        -------
        Union[np.ndarray, pd.Series]
            np.nan where rule is not activated, rule's prediction where it is. If xs vas given and it was a dataframe,
            return a pd.Series. Else, a np.ndarray.
        """
        t0 = time()
        if xs is not None:
            self.calc_activation(xs)
        elif self.activation is None:
            raise ValueError("If the activation vector has not been computed yet, xs can not be None.")
        act = self.activation
        to_ret = np.array([np.nan] * len(act))
        if isinstance(self.prediction, str):
            if self.prediction == "nan":
                raise ValueError("Prediction should not be the 'nan' string, it will conflict with NaNs."
                                 "Rename your class.")
            to_ret = to_ret.astype(str)
        to_ret[act == 1] = self.prediction
        self._time_predict = time() - t0
        if xs is not None and not isinstance(xs, np.ndarray):
            return xs.__class__(index=xs.index, data=to_ret).squeeze()  # So not to requier pandas explicitly
        return to_ret

    def get_correlation(self, other: "Rule") -> float:
        """ Computes the correlation between self and other
        Correlation is the number of points in common between the two vectors divided by their length, times the product
        of the rules' signs.
        Both vectors must have the same length.
        """
        if not len(self) == len(other):
            raise ValueError("Both vectors must have the same length")

        sign = (self.prediction / abs(self.prediction)) * (other.prediction / abs(other.prediction))
        return self._activation.get_correlation(other._activation) * sign


# noinspection PyUnresolvedReferences
class RegressionRule(Rule):

    """Rule applied on continuous target data."""

    def __init__(
        self, condition: Optional[Condition] = None, activation: Optional[Activation] = None,
    ):
        super().__init__(condition, activation)
        self._std = None
        self._criterion = None

        # Inspection / Audit attributs
        self._time_calc_criterion = -1
        self._time_calc_prediction = -1
        self._time_calc_std = -1

    @property
    def std(self) -> float:
        return self._std

    @property
    def criterion(self) -> float:
        return self._criterion

    @property
    def time_calc_prediction(self):
        return self._time_calc_prediction

    @property
    def time_calc_criterion(self):
        return self._time_calc_criterion

    @property
    def time_calc_std(self):
        return self._time_calc_std

    def calc_attributes(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
        """Computes prediction, standard deviation, and regression criterion
        
        Parameters
        ----------
        y: Union[np.ndarray, pd.Series]
            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
            or pd.Series.
        kwargs: dict
            Arguments for calc_regression_criterion
        """
        self.calc_prediction(y)
        self.calc_std(y)
        prediction_vector = self.prediction * self.activation
        self.calc_criterion(prediction_vector, y, **kwargs)

    def calc_prediction(self, y: [np.ndarray, "pd.Series"]):
        """Computes the mean of all activated points in target y and use it as prediction
        
        Parameters
        ----------
        y: [np.ndarray, pd.Series]
            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
            or pd.Series
        """
        t0 = time()
        if self.activation is None:
            return None
        self._prediction = functions.conditional_mean(self.activation, y)
        self._time_calc_prediction = time() - t0
        self.check_thresholds("prediction")

    def calc_std(self, y: Union[np.ndarray, "pd.Series"]):
        """Computes the standard deviation of all activated points in target y
        
        Parameters
        ----------
        y: Union[np.ndarray, pd.Series]
            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
            or pd.Series.
        """
        t0 = time()
        if self.activation is None:
            return None
        self._std = functions.conditional_std(self.activation, y)
        self._time_calc_std = time() - t0
        self.check_thresholds("std")

    def calc_criterion(self, p: Union[np.ndarray, "pd.Series"], y: Union[np.ndarray, "pd.Series"], **kwargs):
        """
        Parameters
        ----------
        p: Union[np.ndarray, pd.Series]
            Prediction vector. Must be a 1-D np.ndarray or pd.Series.
        y: Union[np.ndarray, pd.Series]
            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
            or pd.Series.
        kwargs: dict
            Arguments for calc_regression_criterion
        """
        t0 = time()
        self._criterion = functions.calc_regression_criterion(p, y, **kwargs)
        self._time_calc_criterion = time() - t0
        self.check_thresholds("criterion")


# noinspection PyUnresolvedReferences
class ClassificationRule(Rule):

    """Rule applied on discret target data."""

    def __init__(
        self, condition: Optional[Condition] = None, activation: Optional[Activation] = None,
    ):
        super().__init__(condition, activation)

        self._criterion = None

        self._time_calc_criterion = -1
        self._time_calc_prediction = -1

    @property
    def prediction(self) -> Union[int, str, None]:
        if self._prediction is not None:
            if isinstance(self._prediction, (float, int, str)):
                return self._prediction
            prop = [p[1] for p in self._prediction]
            idx = prop.index(max(prop))
            return self._prediction[idx][0]
        else:
            return None

    @property
    def criterion(self) -> float:
        return self._criterion

    def calc_attributes(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
        """
        Parameters
        ----------
        xs: Union[pd:DataFrame, np.ndarray]
            The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
            or pd:DataFrame.
        y: Union[np.ndarray, pd.Series]
            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
            or pd.Series.
        kwargs: dict
            Arguments for calc_classification_criterion
        """
        self.calc_prediction(y)
        self.calc_criterion(y, **kwargs)

    def calc_prediction(self, y: [np.ndarray, "pd.Series"]):
        """
        Parameters
        ----------
        y: [np.ndarray, pd.Series]
            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
            or pd.Series.
        """
        t0 = time()
        if self.activation is None:
            raise ValueError("The activation vector has not been computed yet.")
        self._prediction = functions.most_common_class(self.activation, y)
        self._time_calc_prediction = time() - t0
        self.check_thresholds("prediction")

    def calc_criterion(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
        """
        Parameters
        ----------
        y: Union[np.ndarray, pd.Series]
            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
            or pd.Series
        kwargs: dict
            Arguments for calc_classification_criterion
        """
        t0 = time()
        self._criterion = functions.calc_classification_criterion(self.activation, self.prediction, y, **kwargs)
        self._time_calc_criterion = time() - t0
        self.check_thresholds("criterion")
